import inspect
from collections.abc import Callable, Iterable, Sequence
from typing import Any, TypeVar

import numpy as np
import pandas as pd

from pyranges.core.names import (
    BY_ENTRY_IN_KWARGS,
    END_COL,
    JOIN_SUFFIX,
    RANGE_COLS,
    SKIP_IF_DF_EMPTY_TYPE,
    SKIP_IF_EMPTY_ANY,
    SKIP_IF_EMPTY_BOTH,
    SKIP_IF_EMPTY_LEFT,
    SKIP_IF_EMPTY_RIGHT,
    START_COL,
    VALID_BY_TYPES,
    VALID_COMBINE_OPTIONS,
    VALID_DIRECTION_TYPE,
    VALID_JOIN_TYPE,
    VALID_OVERLAP_TYPE,
    CombineIntervalColumnsOperation,
)
from pyranges.core.pyranges_helpers import arg_to_list, factorize, factorize_binary
from pyranges.core.tostring import tohtml, tostring
from pyranges.methods.complement_overlaps import _complement_overlaps
from pyranges.methods.join import _both_dfs
from pyranges.methods.merge import _merge
from pyranges.methods.sort import sort_factorize_dict
from pyranges.range_frame.range_frame_validator import InvalidRangesReason


def should_skip_operation(df: pd.DataFrame, *, df2: pd.DataFrame, skip_if_empty: SKIP_IF_DF_EMPTY_TYPE) -> bool:
    """Whether to skip operation because one or more dfs are empty."""
    if df.empty and df2.empty:
        return skip_if_empty in {SKIP_IF_EMPTY_BOTH, SKIP_IF_EMPTY_ANY, SKIP_IF_EMPTY_LEFT, SKIP_IF_EMPTY_RIGHT}
    if df.empty:
        return skip_if_empty in {SKIP_IF_EMPTY_LEFT, SKIP_IF_EMPTY_ANY}
    if df2.empty:
        return skip_if_empty in {SKIP_IF_EMPTY_RIGHT, SKIP_IF_EMPTY_ANY}
    return False


TRangeFrame = TypeVar("TRangeFrame", bound="RangeFrame")


class RangeFrame(pd.DataFrame):
    """Class for range based operations.

    A table with Start and End columns. Parent class of PyRanges. Subclass of pandas DataFrame.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def __str__(
        self,
        **kwargs: int | None,
    ) -> str:  # , max_col_width: int | None = None, max_total_width: int | None = None) -> str:
        """Return string representation."""
        str_repr = tostring(
            self,
            max_col_width=kwargs.get("max_col_width"),
            max_total_width=kwargs.get("max_total_width"),
        )
        if reasons := InvalidRangesReason.formatted_reasons_list(self):
            return f"{str_repr}\nInvalid ranges:\n{reasons}"
        return str_repr

    def __repr__(self, max_col_width: int | None = None, max_total_width: int | None = None) -> str:
        return self.__str__(max_col_width=max_col_width, max_total_width=max_total_width)

    def _repr_html_(self) -> str:
        return tohtml(self)

    def merge_overlaps(
        self,
        *,
        count_col: str | None = None,
        match_by: VALID_BY_TYPES = None,
        slack: int = 0,
    ) -> "RangeFrame":
        """Merge overlapping intervals into one.

        Merge overlapping intervals into a single superinterval by uniting intervals that overlap,
        optionally allowing a small gap (specified by ``slack``) between intervals to be merged. The resulting
        RangeFrame will contain the merged intervals, and if ``count_col`` is provided, a column with the counts
        of merged intervals will be included.

        Parameters
        ----------
        count_col : str or None, default None
            Name of the column to store the count of intervals merged into each superinterval.
            If None, no count column is added.
        match_by : str or list, default None
            Column(s) to group intervals by before merging. Only intervals with equal values in the specified
            column(s) will be considered as overlapping.
        slack : int, default 0
            Allow this many nucleotides between intervals to still consider them overlapping.

        Returns
        -------
        RangeFrame
            A RangeFrame with merged (super) intervals. Metadata columns, index, and order are not necessarily preserved.

        """
        match_by = arg_to_list(match_by)
        return _merge(self, by=match_by, count_col=count_col, slack=slack)

    def count_overlaps(
        self,
        other: "RangeFrame",
        *,
        match_by: str | list[str] | None = None,
        slack: int = 0,
    ) -> "pd.Series":
        """Count the number of overlaps per interval.

        For each interval in self, count how many intervals in ``other`` overlap with it.
        The overlap computation is based on the start and end coordinates, with an optional
        ``slack`` parameter to adjust the overlap threshold by temporarily extending the intervals.

        Parameters
        ----------
        other : RangeFrame
            The RangeFrame whose intervals are compared against those in self for overlap counting.
        match_by : str or list, default None
            Column(s) to group intervals by when determining overlaps. Only intervals with equal values in the specified
            column(s) will be considered as overlapping.
        slack : int, default 0
            Temporarily extend intervals in self by this many nucleotides before checking for overlaps,
            thereby adjusting the overlap threshold.

        Returns
        -------
        pd.Series
            A pandas Series where each element corresponds to the number of overlapping intervals in ``other``
            for the corresponding interval in self.

        """
        import ruranges

        f1, f2 = factorize_binary(self, other, match_by)

        return ruranges.count_overlaps(  # type: ignore[attr-defined]
            self[START_COL].to_numpy(),
            self[END_COL].to_numpy(),
            other[START_COL].to_numpy(),
            other[END_COL].to_numpy(),
            f1,
            f2,
            slack=slack,
        )

    def combine_interval_columns(
        self,
        function: VALID_COMBINE_OPTIONS | CombineIntervalColumnsOperation = "intersect",
        *,
        start: str = START_COL,
        end: str = END_COL,
        start2: str = START_COL + JOIN_SUFFIX,
        end2: str = END_COL + JOIN_SUFFIX,
        drop_old_columns: bool = True,
    ) -> "RangeFrame":
        """Use two pairs of columns representing intervals to create a new start and end column.

        The function is designed as post-processing after join_overlaps to aggregate the coordinates of the two intervals.
        By default, the new start and end columns will be the intersection of the intervals.

        Parameters
        ----------
        function : {"intersect", "union", "swap"} or Callable, default "intersect"
            How to combine the self and other intervals: "intersect", "union", or "swap"
            If a callable is passed, it should take four Series arguments: start1, end1, start2, end2;
            and return a tuple of two integers: (new_starts, new_ends).

        start : str, default "Start"
            Column name for Start of first interval
        end : str, default "End"
            Column name for End of first interval
        start2 : str, default "Start_b"
            Column name for Start of second interval
        end2 : str, default "End_b"
            Column name for End of second interval
        drop_old_columns : bool, default True
            Whether to drop the above mentioned columns.

        """
        from pyranges.methods.combine_positions import (
            _intersect_interval_columns,
            _swap_interval_columns,
            _union_interval_columns,
        )

        if function == "intersect":
            function = _intersect_interval_columns
        elif function == "union":
            function = _union_interval_columns
        elif function == "swap":
            function = _swap_interval_columns

        new_starts, new_ends = function(self[start], self[end], self[start2], self[end2])

        z = self.copy()
        z[START_COL] = new_starts
        z[END_COL] = new_ends

        cols_to_drop = list({start, end, start2, end2}.difference(RANGE_COLS) if drop_old_columns else {})

        return z.drop_and_return(labels=cols_to_drop, axis="columns")

    def cluster_overlaps(
        self,
        *,
        match_by: VALID_BY_TYPES = None,
        cluster_column: str = "Cluster",
        slack: int = 0,
    ) -> "RangeFrame":
        """Give overlapping intervals a common id.

        Parameters
        ----------
        match_by : str or list, default None
            If provided, only intervals with an equal value in column(s) `match_by` may be considered as overlapping.

        slack : int, default 0
            Length by which the criteria of overlap are loosened.
            A value of 1 clusters also bookended intervals.
            Higher slack values cluster more distant intervals (with a maximum distance of slack-1 between them).

        cluster_column:
            Name the cluster column added in output. Default: "Cluster"

        Returns
        -------
        RangeFrame
            RangeFrame with an ID-column "Cluster" added.

        See Also
        --------
        RangeFrame.merge: combine overlapping intervals into one

        """
        import ruranges

        match_by = arg_to_list(match_by)

        factorized = factorize(self, match_by)
        cluster, idx = ruranges.cluster(  # type: ignore[attr-defined]
            self[START_COL].to_numpy(),
            self[END_COL].to_numpy(),
            factorized,
            slack,
        )

        res = self.take(idx).copy()  # type: ignore[arg-type]
        res.insert(res.shape[1], cluster_column, cluster)
        return _mypy_ensure_rangeframe(res)

    # complement_overlaps: unexpected behavior. Not sure what to do with it. Right now, not exposed in PyRanges API.
    def _complement_overlaps(
        self: "RangeFrame",
        other: "RangeFrame",
        *,
        group_by: VALID_BY_TYPES = None,
        slack: int = 0,
    ) -> "RangeFrame":
        """Return the non-overlaps of self with other.

        Parameters
        ----------
        other : PyRanges
           PyRanges to find non-overlaps with.

        group_by : str or list, default None
            Column(s) to group by intervals (i.e. exons). If provided, the complement will be calculated for each group.

        use_strand: {"auto", True, False}, default "auto"
            Whether to return complement separately for intervals on the positive and negative strand.
            The default "auto" means use strand information if present and valid (see .strand_valid)

        chromsizes : dict or PyRanges or pyfaidx.Fasta
            If provided, external complement intervals will also be returned, i.e. the intervals corresponding to the
            beginning of the chromosome up to the first interval and from the last interval to the end of the
            chromosome. If group_by is provided, these are returned for each group.
            Format of chromsizes: dict or PyRanges describing the lengths of the chromosomes.
            pyfaidx.Fasta object is also accepted since it conveniently loads chromosome length

        slack : int, default 0
            An integer offset that adjusts the overlap threshold when computing the complement intervals.
            Negative values reduce the required gap between intervals (effectively "shrinking" them), while positive values
            increase the gap threshold.

        Returns
        -------
        PyRanges
            Non-overlapping intervals

        Notes
        -----
        * To ensure non-overlap among the input intervals, merge_overlaps is run before the complement is calculated.
        * Bookended intervals will result in no complement intervals returned since they would be of length 0.

        See Also
        --------
        PyRanges.subtract_overlaps : report non-overlapping subintervals
        PyRanges.outer_ranges : report the boundaries of groups of intervals (e.g. transcripts/genes)

        """
        group_by = arg_to_list(group_by)
        return _complement_overlaps(self, other, by=group_by, slack=slack)

    def join_overlaps(
        self,
        other: "RangeFrame",
        *,
        join_type: VALID_JOIN_TYPE = "inner",
        multiple: VALID_OVERLAP_TYPE = "all",
        match_by: VALID_BY_TYPES = None,
        slack: int = 0,
        suffix: str = JOIN_SUFFIX,
        contained_intervals_only: bool = False,
        report_overlap_column: str | None = None,
    ) -> "RangeFrame":
        """Join RangeFrame objects based on overlapping intervals.

        Find pairs of overlapping intervals between self and other and combine their attributes.
        Each row in the output contains columns from both intervals, including their start and end positions.
        By default, only overlapping intervals are included, but the join_type parameter controls how intervals
        without overlaps are handled.

        Parameters
        ----------
        other : RangeFrame
            The RangeFrame to join with.
        join_type : {"inner", "left", "right", "outer"}, default "inner"
            Specifies how to handle intervals that do not overlap. "inner" returns only overlapping intervals,
            "left" returns all intervals from self (with missing values for non-overlapping intervals from other),
            "right" returns all intervals from other, and "outer" returns all intervals from both.
        multiple : {"all", "first", "last"}, default "all"
            Determines which overlapping interval(s) to report when multiple intervals in other overlap the same interval in self.
            "all" reports all overlaps (which may lead to duplicate rows), "first" reports only the overlapping interval with
            the smallest start in other, and "last" reports only the overlapping interval with the largest end in other.
        match_by : str or list, default None
            If provided, only intervals with matching values in the specified column(s) will be joined.
        slack : int, default 0
            Temporarily extend intervals in self by this many units on both ends before checking for overlaps.
        suffix : str, default JOIN_SUFFIX
            Suffix to append to columns from the other RangeFrame in the output.
        contained_intervals_only : bool, default False
            If True, only join intervals from self that are entirely contained within an interval from other.
        report_overlap_column : str or None, default None
            If provided, add a column with this name reporting the amount of overlap between joined intervals.
            The overlap is computed as the minimum of the end positions minus the maximum of the start positions.

        Returns
        -------
        RangeFrame
            A new RangeFrame containing the joined intervals with columns from both input RangeFrames.
            The indices of the input RangeFrames are not preserved in the output.

        Notes
        -----
        Attributes from the other RangeFrame may have their column names modified by appending the specified suffix.

        """
        res = _both_dfs(
            self,
            other,
            by=match_by,
            slack=slack,
            multiple=multiple,
            contained=contained_intervals_only,
            join_type=join_type,
            suffix=suffix,
        )

        if report_overlap_column:
            res[report_overlap_column] = res[["End", "End" + suffix]].min(axis=1) - res[
                ["Start", "Start" + suffix]
            ].max(axis=1)

        res.index = res.index.astype(np.int64)

        return _mypy_ensure_rangeframe(res)

    def max_disjoint_overlaps(
        self,
        *,
        slack: int = 0,
        match_by: VALID_BY_TYPES = None,
    ) -> "RangeFrame":
        """Find the maximal disjoint set of intervals.

        Returns a subset of the rows in self so that no two intervals overlap, choosing those that
        maximize the number of intervals in the result.

        Parameters
        ----------
        slack : int, default 0
            Length by which the criteria of overlap are loosened.
            A value of 1 implies that bookended intervals are considered overlapping.
            Higher slack values allow more distant intervals (with a maximum distance of slack-1 between them).

        match_by : str or list, default None
            If provided, only intervals with an equal value in column(s) `match_by` may be considered as overlapping.

        Returns
        -------
        RangeFrame
            RangeFrame with maximal disjoint set of intervals.

        See Also
        --------
        RangeFrame.merge_overlaps : merge intervals into non-overlapping superintervals
        RangeFrame.cluster : annotate overlapping intervals with common ID

        """
        import ruranges

        factorized = factorize(self, match_by)

        idx = ruranges.max_disjoint(  # type: ignore[attr-defined]
            starts=self[START_COL].to_numpy(),
            ends=self[END_COL].to_numpy(),
            groups=factorized,
            slack=slack,
        )
        return _mypy_ensure_rangeframe(self.take(idx))  # type: ignore[arg-type]

    def nearest_ranges(
        self,
        other: "RangeFrame",
        *,
        match_by: VALID_BY_TYPES = None,
        suffix: str = JOIN_SUFFIX,
        exclude_overlaps: bool = False,
        k: int = 1,
        dist_col: str | None = "Distance",
        direction: VALID_DIRECTION_TYPE = "any",
    ) -> "RangeFrame":
        """Find closest interval.

        For each interval in self RangeFrame, the columns of the nearest interval in other RangeFrame are appended.

        Parameters
        ----------
        other : RangeFrame
            RangeFrame to find nearest interval in.

        exclude_overlaps : bool, default True
            Whether to not report intervals of others that overlap with self as the nearest ones.

        direction : {"any", "forward", "backward"}, default "any", i.e. both directions
            Whether to only look for nearest in one direction.

        match_by : str or list, default None
            If provided, only intervals with an equal value in column(s) `match_by` may be matched.

        k : int, default 1
            Number of nearest intervals to fetch.

        suffix : str, default "_b"
            Suffix to give columns with shared name in other.

        dist_col : str or None
            Optional column to store the distance in.

        Returns
        -------
        RangeFrame

            A RangeFrame with columns representing nearest interval horizontally appended.

        See Also
        --------
        RangeFrame.join_overlaps : Has a slack argument to find intervals within a distance.

        """
        import ruranges

        f1, f2 = factorize_binary(self, other, match_by)
        idx1, idx2, dist = ruranges.nearest(  # type: ignore[attr-defined]
            groups=f1,
            starts=self[START_COL].to_numpy(),
            ends=self[END_COL].to_numpy(),
            groups2=f2,
            starts2=other[START_COL].to_numpy(),
            ends2=other[END_COL].to_numpy(),
            k=k,
            slack=0,
            include_overlaps=not exclude_overlaps,
            direction=direction,
        )

        left = self.take(idx1)  # type: ignore[arg-type]
        to_concat: list[pd.DataFrame | pd.Series] = [
            left.reset_index(drop=True),
            pd.DataFrame(other).take(idx2).add_suffix(suffix).reset_index(drop=True),  # type: ignore[arg-type]
        ]
        if dist_col is not None:
            to_concat.append(pd.Series(dist, name=dist_col))

        res = pd.concat(to_concat, axis=1)
        res.index = pd.Index(left.index.to_numpy())

        return _mypy_ensure_rangeframe(res)

    def overlap(
        self,
        other: "RangeFrame",
        multiple: VALID_OVERLAP_TYPE = "all",
        slack: int = 0,
        *,
        contained_intervals_only: bool = False,
        match_by: VALID_BY_TYPES = None,
    ) -> "RangeFrame":
        """Return overlapping intervals.

        Returns the intervals in self which overlap with those in other.

        Parameters
        ----------
        other : RangeFrame
            RangeFrame to find overlaps with.

        multiple : {"all", "first", "last"}, default "all"
            What intervals to report when multiple intervals in 'other' overlap with the same interval in self.
            The default "all" reports all overlapping subintervals, which will have duplicate indices.
            "first" reports only, for each interval in self, the overlapping subinterval with smallest Start in 'other'
            "last" reports only the overlapping subinterval with the biggest End in 'other'

        slack : int, default 0
            Intervals in self are temporarily extended by slack on both ends before overlap is calculated, so that
            we allow non-overlapping intervals to be considered overlapping if they are within less than slack distance
            e.g. slack=1 reports bookended intervals.

        contained_intervals_only : bool, default False
            Whether to report only intervals that are entirely contained in an interval of 'other'.

        match_by : str or list, default None
            If provided, only overlapping intervals with an equal value in column(s) `match_by` are reported.

        Returns
        -------
        RangeFrame

            A RangeFrame with overlapping intervals.

        See Also
        --------
        RangeFrame.intersect : report overlapping subintervals
        RangeFrame.set_intersect : set-intersect RangeFrame (e.g. merge then intersect)

        """
        from pyranges.methods.overlap import _overlap

        by = arg_to_list(match_by)

        result = _overlap(
            self,
            other,
            by=by,
            slack=slack,
            multiple=multiple,
            contained=contained_intervals_only,
        )

        return _mypy_ensure_rangeframe(result)

    def sort_ranges(
        self: "RangeFrame",
        by: VALID_BY_TYPES = None,
        *,
        natsort: bool = True,
        sort_rows_reverse_order: Sequence[bool] | None = None,
    ) -> "RangeFrame":
        """Sort RangeFrame according to Start, End, and any other columns given.

        For uses not covered by this function, use  DataFrame.sort_values().

        Parameters
        ----------
        by : str or list of str, default None
            in the desired order as part of the 'by' argument.

        natsort : bool, default False
            Whether to use natural sorting for the columns in match_by.

        sort_rows_reverse_order : sequence of bools or None
            Whether to sort these rows in the reverse order for the starts and ends.

        Returns
        -------
        RangeFrame

            Sorted RangeFrame. The index is preserved. Use .reset_index(drop=True) to reset the index.

        """
        import ruranges

        by = arg_to_list(by)
        by_sort_order_as_int = sort_factorize_dict(self, by, use_natsort=natsort)
        idxs = ruranges.sort_intervals(  # type: ignore[attr-defined]
            by_sort_order_as_int,
            self[START_COL].to_numpy(),
            self[END_COL].to_numpy(),
            sort_reverse_direction=np.array(sort_rows_reverse_order, dtype=bool) if sort_rows_reverse_order else None,
        )
        return _mypy_ensure_rangeframe(self.take(idxs))  # type: ignore[arg-type]

    def subtract_overlaps(
        self: "RangeFrame",
        other: "RangeFrame",
        match_by: VALID_BY_TYPES = None,
    ) -> "RangeFrame":
        """Subtract intervals, i.e. return non-overlapping subintervals.

        Identify intervals in other that overlap with intervals in self; return self with the overlapping parts removed.

        Parameters
        ----------
        other:
            RangeFrame to subtract.

        match_by : str or list, default None
            If provided, only intervals with an equal value in column(s) `match_by` may be considered as overlapping.

        Returns
        -------
        RangeFrame
            RangeFrame with subintervals from self that do not overlap with any interval in other.
            Columns and index are preserved.

        Warning
        -------
        The returned Pyranges may have index duplicates. Call .reset_index(drop=True) to fix it.

        See Also
        --------
        RangeFrame.overlap : use with invert=True to return all intervals without overlap
        RangeFrame.complement_ranges : return the internal complement_ranges of intervals, i.e. its introns.

        """
        import ruranges

        f1, f2 = factorize_binary(self, other, match_by)

        idx, start, end = ruranges.subtract(  # type: ignore[attr-defined]
            self[START_COL].to_numpy(),
            self[END_COL].to_numpy(),
            other[START_COL].to_numpy(),
            other[END_COL].to_numpy(),
            f1,
            f2,
        )

        output = self.take(idx).copy()  # type: ignore[arg-type]
        output[START_COL], output[END_COL] = start, end
        return _mypy_ensure_rangeframe(output)

    def sort_by_position(self) -> "RangeFrame":
        """Sort by Start and End columns."""
        return _mypy_ensure_rangeframe(self.sort_values(RANGE_COLS))

    def reasons_why_frame_is_invalid(self) -> list[InvalidRangesReason] | None:  # noqa: D102
        __doc__ = InvalidRangesReason.is_invalid_ranges_reasons.__doc__  # noqa: F841

        return InvalidRangesReason.is_invalid_ranges_reasons(self)

    def copy(self, *args, **kwargs) -> "RangeFrame":  # noqa: D102
        return _mypy_ensure_rangeframe(super().copy(*args, **kwargs))

    def drop(self, *args, **kwargs) -> "RangeFrame | None":  # type: ignore[override]  # noqa: D102
        return self.__class__(super().drop(*args, **kwargs))

    def drop_and_return[T: "RangeFrame"](self: T, *args: Any, **kwargs: Any) -> T:  # noqa: PYI019, D102
        kwargs["inplace"] = False
        return self.__class__(super().drop(*args, **kwargs))

    def reindex(self, *args, **kwargs) -> "RangeFrame":  # noqa: D102
        return self.__class__(super().reindex(*args, **kwargs))


def _mypy_ensure_rangeframe(r: pd.DataFrame) -> "RangeFrame":
    result = RangeFrame(r)
    if not isinstance(result, RangeFrame):
        msg = f"Expected RangeFrame, got {type(result)}"
        raise TypeError(msg)
    return result


def assert_valid_ranges(function: Callable, *args: "RangeFrame") -> None:
    """Raise ValueError because function is called on invalid ranges."""
    if any(r.reasons_why_frame_is_invalid() for r in args):
        is_not_lambda = function.__name__ != "<lambda>"
        function_repr = function.__name__ if is_not_lambda else inspect.getsource(function).strip()
        msg = f"Cannot perform function on invalid ranges (function was {function_repr})."
        raise ValueError(msg)


def _with_group_keys_to_kwargs(by: list[str]) -> Callable:
    def decorator(func) -> Callable:
        def wrapper(group, **kwargs) -> Any:
            # Extract the group keys and add them to kwargs
            names = group.name if isinstance(group.name, Iterable) and not isinstance(group.name, str) else [group.name]
            kwargs[BY_ENTRY_IN_KWARGS] = dict(zip(by, names, strict=True))
            return func(group, **kwargs)

        return wrapper

    return decorator



import numpy as np
import pandas as pd


def _intersect_interval_columns(
    starts: pd.Series,
    ends: pd.Series,
    starts2: pd.Series,
    ends2: pd.Series,
) -> tuple[pd.Series, pd.Series]:
    new_starts = pd.Series(
        np.where(starts > starts2.to_numpy(), starts, starts2),
        index=starts.index,
    )
    new_ends = pd.Series(
        np.where(ends < ends2.to_numpy(), ends, ends2),
        index=ends.index,
    )
    return new_starts, new_ends


def _union_interval_columns(
    starts: pd.Series,
    ends: pd.Series,
    starts2: pd.Series,
    ends2: pd.Series,
) -> tuple[pd.Series, pd.Series]:
    new_starts = pd.Series(
        np.where(starts < starts2.to_numpy(), starts, starts2),
        index=starts.index,
    )
    new_ends = pd.Series(
        np.where(ends > ends2.to_numpy(), ends, ends2),
        index=ends.index,
    )
    return new_starts, new_ends


def _swap_interval_columns(
    starts: pd.Series,  # noqa: ARG001
    ends: pd.Series,  # noqa: ARG001
    starts2: pd.Series,
    ends2: pd.Series,
) -> tuple[pd.Series, pd.Series]:
    return starts2, ends2



# must import here every module/object X that will be available as pyranges.X
# For each 'extension' e.g. named foobar, you must have:
# a ext/foobar.py file in pyranges/ which contains the actual code for methods etc
# a foobar.py file in pyranges/ which just imports the minimal objects to be exposed

from pyranges import orfs, seqs, stats  # noqa: F401
from pyranges.core.assistant import assistant  # noqa: F401
from pyranges.core.example_data import example_data  # noqa: F401
from pyranges.core.multioverlap import count_overlaps  # noqa: F401
from pyranges.core.options import option_manager as options  # noqa: F401
from pyranges.core.pyranges_main import PyRanges  # noqa: F401
from pyranges.core.random import random  # noqa: F401
from pyranges.core.version import __version__  # noqa: F401
from pyranges.methods.concat import concat  # noqa: F401
from pyranges.methods.tile_genome import tile_genome  # noqa: F401
from pyranges.range_frame.range_frame import RangeFrame  # noqa: F401
from pyranges.readers import from_string, read_bam, read_bed, read_bigwig, read_gff3, read_gtf  # noqa: F401

read_gff = read_gtf



import subprocess
import tempfile
from io import StringIO

import hypothesis.strategies as st
import numpy as np
import pandas as pd
import pytest
from hypothesis import given, settings
from hypothesis.extra.pandas import column, data_frames, indexes

from pyranges import PyRanges
from pyranges.core.names import VALID_GENOMIC_STRAND_INFO

max_examples = 15
slow_max_examples = 5
deadline = None

starts = st.integers(min_value=1, max_value=int(1e7))
lengths = st.integers(min_value=1, max_value=int(1e4))

strands = st.sampled_from(VALID_GENOMIC_STRAND_INFO)

chromosomes = st.sampled_from([f"chr{e!s}" for e in [*list(range(1, 23)), "X", "Y", "M"]])


dfs_zero_length_not_allowed = data_frames(
    index=indexes(dtype=np.int64, min_size=1, unique=True, elements=starts),
    columns=[
        column("Chromosome", chromosomes),
        column("Start", elements=starts),
        column("End", elements=lengths),
        column("Strand", strands),
    ],
)


@st.composite
def nonempty_pyranges(draw):  # nosec
    df = draw(dfs_zero_length_not_allowed)
    df.loc[:, "End"] += df.Start

    df.insert(3, "Name", "a")
    df.insert(4, "Score", 0)

    return PyRanges(df)


def run_bedtools(command, gr, gr2, strand_behavior, nearest_overlap=False, nearest_how=None, ties=""):
    bedtools_strand_behavior = {"ignore": "", "same": "-s", "opposite": "-S"}[strand_behavior]
    bedtools_overlap = {True: "", False: "-io"}[nearest_overlap]
    bedtools_how = {"upstream": "-id", "downstream": "-iu", None: ""}[nearest_how] + " -D a"
    ties = "-t " + ties if ties else ""

    # file1_proc_sub = f"""<(echo -e '{gr.to_csv(sep="\t", header=False, index=False)}')"""
    # file2_proc_sub = f"""<(echo -e '{gr2.to_csv(sep="\t", header=False, index=False)}')"""

    # cmd = command.format(
    #     f1=file1_proc_sub,
    #     f2=file2_proc_sub,
    #     strand=bedtools_strand_behavior,
    #     overlap=bedtools_overlap,
    #     bedtools_how=bedtools_how,
    #     ties=ties,
    # )
    # return subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()  # nosec  # nosec

    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = f"{temp_dir}/f1.bed"
        f2 = f"{temp_dir}/f2.bed"
        gr.to_csv(f1, sep="\t", header=False, index=False)
        gr2.to_csv(f2, sep="\t", header=False, index=False)

        cmd = command.format(
            f1=f1,
            f2=f2,
            strand=bedtools_strand_behavior,
            overlap=bedtools_overlap,
            bedtools_how=bedtools_how,
            ties=ties,
        )
        # ignoring the below line in bandit as only strings created by
        # the test suite is run here; no user input ever sought
        return subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()  # nosec  # nosec


def assert_equal(result, bedtools_df) -> None:
    if result.empty and bedtools_df.empty:
        return
    result = PyRanges(result).sort_ranges().reset_index(drop=True)
    bedtools_df = PyRanges(bedtools_df).sort_ranges().reset_index(drop=True)
    pd.testing.assert_frame_equal(result, bedtools_df, check_exact=False, atol=1e-5)


strand_behavior = ["ignore", "same", "opposite"]


@pytest.mark.bedtools
@pytest.mark.parametrize("strand_behavior", strand_behavior)
@settings(
    max_examples=max_examples,
    print_blob=True,
)
@given(gr=nonempty_pyranges(), gr2=nonempty_pyranges())  # pylint: disable=no-value-for-parameter
def test_overlap(gr, gr2, strand_behavior) -> None:
    overlap_command = "bedtools intersect -u {strand} -a {f1} -b {f2}"

    bedtools_result = run_bedtools(overlap_command, gr, gr2, strand_behavior)

    bedtools_df = pd.read_csv(
        StringIO(bedtools_result),
        header=None,
        names=["Chromosome", "Start", "End", "Name", "Score", "Strand"],
        sep="\t",
    )

    result = gr.overlap(gr2, strand_behavior=strand_behavior)

    result = result.reset_index(drop=True)
    if result.empty and bedtools_df.empty:
        return
    assert_equal(result, bedtools_df)


@pytest.mark.bedtools
@pytest.mark.parametrize("strand_behavior", strand_behavior)
@settings(
    max_examples=max_examples,
    print_blob=True,
)
@given(gr=nonempty_pyranges(), gr2=nonempty_pyranges())  # pylint: disable=no-value-for-parameter
def test_coverage(gr, gr2, strand_behavior) -> None:
    coverage_command = "bedtools coverage {strand} -a {f1} -b {f2}"

    bedtools_result = run_bedtools(coverage_command, gr, gr2, strand_behavior)

    bedtools_df = PyRanges(
        pd.read_csv(
            StringIO(bedtools_result),
            header=None,
            usecols=[0, 1, 2, 3, 4, 5, 6, 9],
            names=["Chromosome", "Start", "End", "Name", "Score", "Strand", "NumberOverlaps", "FractionOverlaps"],
            dtype={"FractionOverlap": np.float64},
            sep="\t",
        )
    )

    result = gr.coverage(gr2, strand_behavior=strand_behavior)

    assert_equal(result, bedtools_df)


STRAND_BEHAVIOR_NO_OPPOSITE = ["ignore", "same"]


@pytest.mark.bedtools
@pytest.mark.parametrize("strand_behavior", STRAND_BEHAVIOR_NO_OPPOSITE)
@settings(
    max_examples=max_examples,
    print_blob=True,
)
@given(gr=nonempty_pyranges(), gr2=nonempty_pyranges())  # pylint: disable=no-value-for-parameter
def test_set_intersect(gr, gr2, strand_behavior) -> None:
    set_intersect_command = "bedtools intersect {strand} -a <(sort -k1,1 -k2,2n {f1} | bedtools merge {strand} -c 4,5,6 -o first -i -) -b <(sort -k1,1 -k2,2n {f2} | bedtools merge {strand} -c 4,5,6 -o first -i -)"
    bedtools_result = run_bedtools(set_intersect_command, gr, gr2, strand_behavior)

    bedtools_df = read_bedtools_result_set_op(bedtools_result, strand_behavior=strand_behavior)

    result = gr.set_intersect_overlaps(gr2, strand_behavior=strand_behavior)

    if result.empty and bedtools_df.empty:
        return

    assert_equal(result, bedtools_df)


@pytest.mark.bedtools
@pytest.mark.parametrize("strand_behavior", STRAND_BEHAVIOR_NO_OPPOSITE)
@settings(
    max_examples=max_examples,
    deadline=deadline,
    print_blob=True,
)
@given(gr=nonempty_pyranges(), gr2=nonempty_pyranges())
def test_set_union(gr, gr2, strand_behavior) -> None:
    set_union_command = "cat {f1} {f2} | bedtools sort | bedtools merge {strand} -c 4,5,6 -o first -i -"  # set_union_command = "bedtools merge {strand} -c 4,5,6 -o first -i {f1}"
    bedtools_result = run_bedtools(set_union_command, gr, gr2, strand_behavior)

    bedtools_df = read_bedtools_result_set_op(bedtools_result, strand_behavior)

    result = gr.set_union_overlaps(gr2, strand_behavior=strand_behavior)

    assert_equal(bedtools_df, result)


# @pytest.mark.bedtools
# @pytest.mark.parametrize(("nearest_how", "overlap", "strand_behavior"), product(nearest_hows, overlaps, strand_behavior))
# @settings(
#     max_examples=max_examples,
#     deadline=deadline,
#     print_blob=True,
# )
# @given(gr=nonempty_pyranges(), gr2=nonempty_pyranges())  # pylint: disable=no-value-for-parameter
# def test_nearest(gr, gr2, nearest_how, overlap, strand_behavior) -> None:
#     nearest_command = "bedtools closest {bedtools_how} {strand} {overlap} -t first -d -a <(sort -k1,1 -k2,2n {f1}) -b <(sort -k1,1 -k2,2n {f2})"
#
#     bedtools_result = run_bedtools(nearest_command, gr, gr2, strand_behavior, overlap, nearest_how)
#
#     bedtools_df = pd.read_csv(
#         StringIO(bedtools_result),
#         header=None,
#         names="Chromosome Start End Strand Chromosome2 Distance".split(),
#         usecols=[0, 1, 2, 5, 6, 12],
#         sep="\t",
#     )
#
#     bedtools_df.Distance = bedtools_df.Distance.abs()
#
#     bedtools_df = bedtools_df[bedtools_df.Chromosome2 != "."]
#     bedtools_df = bedtools_df.drop("Chromosome2", axis=1)
#
#     result = gr.nearest_ranges(gr2, strand_behavior=strand_behavior, overlap=overlap, how=nearest_how)
#
#     print("bedtools " * 5)
#     print(bedtools_df)
#     print("result " * 5)
#     print(result)
#
#     compare_results_nearest(bedtools_df, result)


@pytest.mark.bedtools
@pytest.mark.parametrize("strand_behavior", strand_behavior)
@settings(
    max_examples=max_examples,
    print_blob=True,
)
@given(gr=nonempty_pyranges(), gr2=nonempty_pyranges())  # pylint: disable=no-value-for-parameter
def test_subtraction(gr, gr2, strand_behavior) -> None:
    subtract_command = "bedtools subtract {strand} -a {f1} -b {f2}"

    bedtools_result = run_bedtools(subtract_command, gr, gr2, strand_behavior)

    bedtools_df = pd.read_csv(
        StringIO(bedtools_result),
        header=None,
        names=["Chromosome", "Start", "End", "Name", "Score", "Strand"],
        sep="\t",
    )

    result = gr.subtract_overlaps(gr2, strand_behavior=strand_behavior)

    assert_equal(result, bedtools_df)


def read_bedtools_result_set_op(bedtools_result, strand_behavior):
    if strand_behavior == "same":
        usecols = [0, 1, 2, 5]
        names = ["Chromosome", "Start", "End", "Strand"]
    elif strand_behavior == "ignore":
        usecols = [0, 1, 2]
        names = ["Chromosome", "Start", "End"]
    else:
        raise ValueError("Invalid strand behavior: " + strand_behavior)

    return pd.read_csv(
        StringIO(bedtools_result),
        header=None,
        usecols=usecols,
        names=names,
        sep="\t",
    )



info = """Utilities to instruct a AI coding assistant for pyranges prompts.

Get a prompt to copy-paste into an AI assistant to prime it for pyranges coding tasks:
    >>> import pyranges as pr
    >>> pr.assistant.prompt()

Make a file with pyranges documentation to upload to the AI assistant:
    >>> pr.assistant.export_docs("pr_docs.txt")"""

prompt_default = """Act as an expert bioinformatician programmer experienced in pyranges (complete documentation attached for you to learn). Next, answer my requests for code by first explaining the workflow, followed by oneliner-style code snippets, as concise as possible but elegant, preceded by the text of task as commented code. Ensure you use pyranges v1 interface that you find here, rather than the v0, from which you may have seen examples before; v1 renamed many methods. """
prompt_add_concise = """Output code that is as concise as possible but elegant. Assume pyranges is fully installed. No import statements. Use aptly named variables, no need to declare them. """


class Assistant:
    """A class to instruct a AI coding assistant for pyranges prompts."""

    def __init__(self) -> None:
        pass

    def __str__(self) -> str:
        return info

    def __repr__(self) -> str:
        return self.__str__()

    def prompt(self, to_file=None, *, concise: bool = False) -> str | None:
        """Get an example prompt to use for the AI coding assistant.

        Parameters
        ----------
        to_file : str or pathlib.Path, optional
            If provided, the prompt will be written to this file instead of returning it as a string.

        concise : bool, default False
            If True, the prompt will request concise code snippets.

        Returns
        -------
        str
            The prompt string to use with the AI coding assistant.

        """
        msg = prompt_default + prompt_add_concise if concise else prompt_default
        if to_file is not None:
            from pathlib import Path

            path = Path(to_file)
            path.write_text(msg, encoding="utf-8")
            return None
        return msg

    def export_docs(self, to_file=None, *, include_df=False) -> str | None:
        """Build a single string containing all RST sources **plus** all public PyRanges docstrings.

        Parameters
        ----------
        to_file : str or pathlib.Path, optional
            Destination path.  If *None* (default) the text is returned.
        include_df : bool, default False
            Also include methods inherited unchanged from :class:`pandas.DataFrame`.

        Returns
        -------
        str | None
            The documentation blob, or None if written to a file.

        Examples
        --------
        >>> import pyranges
        >>> pyranges.assistant.export_docs(to_file="pr_docs.txt")

        """
        from pyranges.methods.export_docs import _export_docs

        return _export_docs(to_file=to_file, include_df=include_df)


assistant = Assistant()



from typing import TYPE_CHECKING, Optional

from pyranges.core.names import VALID_STRAND_BEHAVIOR_TYPE
from pyranges.core.pyranges_helpers import ensure_pyranges
from pyranges.methods import concat

if TYPE_CHECKING:
    from pyranges.core.pyranges_main import PyRanges


def count_overlaps(
    grs: dict[str, "PyRanges"],
    features: Optional["PyRanges"] = None,
    strand_behavior: VALID_STRAND_BEHAVIOR_TYPE = "auto",
    by: list[str] | None = None,
) -> "PyRanges":
    """Count overlaps in multiple pyranges.

    Parameters
    ----------
    grs : dict of PyRanges
        The PyRanges to use as queries.

    features : PyRanges, default None
        The PyRanges to use as subject in the query. If None, the PyRanges themselves are used as a query.

    strand_behavior : {None, "same", "opposite", False}, default None, i.e. auto
        Whether to compare PyRanges on the same strand, the opposite or ignore strand
        information. The default, None, means use "same" if both PyRanges are stranded,
        otherwise ignore the strand information.

     how : {None, "all", "containment", "first"}, default None, i.e. all
        What intervals to report. By default reports all overlapping intervals. "containment"
        reports intervals where the overlapping is contained within it.

    by : list of str, default None
        Columns to group by.

    Examples
    --------
    >>> import pyranges as pr
    >>> a = '''Chromosome Start End
    ... chr1    6    12
    ... chr1    10    20
    ... chr1    22    27
    ... chr1    24    30'''

    >>> b = '''Chromosome Start End
    ... chr1    12    32
    ... chr1    14    30'''

    >>> c = '''Chromosome Start End
    ... chr1    8    15
    ... chr1    10    14
    ... chr1    32    34'''

    >>> grs = {n: pr.from_string(s) for n, s in zip(["a", "b", "c"], [a, b, c])}
    >>> for k, v in grs.items():
    ...     print("Name: " + k)
    ...     print(v)
    Name: a
      index  |    Chromosome      Start      End
      int64  |    object          int64    int64
    -------  ---  ------------  -------  -------
          0  |    chr1                6       12
          1  |    chr1               10       20
          2  |    chr1               22       27
          3  |    chr1               24       30
    PyRanges with 4 rows, 3 columns, and 1 index columns.
    Contains 1 chromosomes.
    Name: b
      index  |    Chromosome      Start      End
      int64  |    object          int64    int64
    -------  ---  ------------  -------  -------
          0  |    chr1               12       32
          1  |    chr1               14       30
    PyRanges with 2 rows, 3 columns, and 1 index columns.
    Contains 1 chromosomes.
    Name: c
      index  |    Chromosome      Start      End
      int64  |    object          int64    int64
    -------  ---  ------------  -------  -------
          0  |    chr1                8       15
          1  |    chr1               10       14
          2  |    chr1               32       34
    PyRanges with 3 rows, 3 columns, and 1 index columns.
    Contains 1 chromosomes.

    >>> pr.count_overlaps(grs)
    index    |    Chromosome    Start    End      a        b        c
    int64    |    object        int64    int64    int64    int64    int64
    -------  ---  ------------  -------  -------  -------  -------  -------
    0        |    chr1          6        8        1        0        0
    1        |    chr1          8        10       1        0        1
    2        |    chr1          10       12       2        0        2
    3        |    chr1          12       14       1        1        2
    ...      |    ...           ...      ...      ...      ...      ...
    8        |    chr1          24       27       2        2        0
    9        |    chr1          27       30       1        2        0
    10       |    chr1          30       32       0        1        0
    11       |    chr1          32       34       0        0        1
    PyRanges with 12 rows, 6 columns, and 1 index columns.
    Contains 1 chromosomes.

    >>> gr = pr.PyRanges({"Chromosome": ["chr1"] * 2, "Start": [0, 25], "End": [40, 35]}).tile_ranges(10)
    >>> gr
      index  |    Chromosome      Start      End
      int64  |    object          int64    int64
    -------  ---  ------------  -------  -------
          0  |    chr1                0       10
          0  |    chr1               10       20
          0  |    chr1               20       30
          0  |    chr1               30       40
          1  |    chr1               20       30
          1  |    chr1               30       40
    PyRanges with 6 rows, 3 columns, and 1 index columns (with 4 index duplicates).
    Contains 1 chromosomes.

    >>> pr.count_overlaps(grs, gr)
      index  |    Chromosome      Start      End        a        b        c
      int64  |    object          int64    int64    int64    int64    int64
    -------  ---  ------------  -------  -------  -------  -------  -------
          0  |    chr1                0       10        1        0        1
          0  |    chr1               10       20        2        2        2
          0  |    chr1               20       30        2        2        0
          0  |    chr1               30       40        0        1        1
          1  |    chr1               20       30        2        2        0
          1  |    chr1               30       40        0        1        1
    PyRanges with 6 rows, 6 columns, and 1 index columns (with 4 index duplicates).
    Contains 1 chromosomes.

    """
    concated = concat.concat(grs.values())
    _features = concated.split_overlaps(between=True) if features is None else features.copy()

    for name, gr in grs.items():
        counts = _features._count_overlaps(  # noqa: SLF001
            gr.remove_nonloc_columns(),
            match_by=by,
            strand_behavior=strand_behavior,
        )
        _features.insert(_features.shape[1], name, counts)
    return ensure_pyranges(_features.astype(dict.fromkeys(grs, int)))



import shutil
from dataclasses import dataclass
from typing import TYPE_CHECKING

import pandas as pd
from tabulate import tabulate

if TYPE_CHECKING:
    from collections.abc import Callable

    import pyranges as pr

import html as _html
import re

from pyranges import options
from pyranges.range_frame.range_frame_validator import InvalidRangesReason


@dataclass
class AdjustedTableData:
    truncated_data: list[list[str]]
    truncated_headers: list[str]
    truncated_dtypes: list[str]
    included_columns: int


def console_width(max_total_width: int | None = None) -> int:
    """Return console width."""
    if max_total_width is not None:
        return max_total_width
    if width := options.get_option("console_width"):
        return width
    return shutil.get_terminal_size().columns


def _find_aliases_for_columns(df: "pr.RangeFrame") -> dict[str, str]:
    import random

    overlapping_cols = _get_indices_with_same_name_as_columns(df)
    return {c: "".join(random.sample(str(c.__hash__()), len(c))) for c in overlapping_cols}


def _get_indices_with_same_name_as_columns(df: "pr.RangeFrame") -> list[str]:
    if not df.index.name and not any(df.index.names):
        return []

    columns = {*df.columns}
    if df.index.name and df.index.name in columns:
        return [str(df.index.name)]

    if cols_in_both := columns.intersection(df.index.names):
        return [*cols_in_both]

    return []


def tostring(
    self: "pr.RangeFrame",
    max_col_width: int | None = None,
    max_total_width: int | None = None,
) -> str:
    """Return string representation."""
    number_index_levels = self.index.nlevels
    number_duplicated_indices = self.index.duplicated().sum()
    has_duplicated_index = number_duplicated_indices > 0

    from pyranges import options

    max_rows_to_show = options.get_option("max_rows_to_show")
    max_column_names_to_show = options.get_option("max_column_names_to_show")

    half_of_max_rows_to_show = max_rows_to_show // 2
    _self = (
        pd.concat([self.head(half_of_max_rows_to_show), self.tail(half_of_max_rows_to_show + 1)])
        if len(self) > max_rows_to_show
        else pd.DataFrame(self.copy())
    )

    replacement_names = _find_aliases_for_columns(self)
    inverted_replacement_names = {v: k for k, v in replacement_names.items()}
    _self.columns = [replacement_names.get(c, c) for c in _self.columns]
    _self = pd.DataFrame(_self).reset_index()

    truncation_marker = ["..."]
    if len(_self) > max_rows_to_show:
        head = [list(v) for _, v in _self.head(half_of_max_rows_to_show).iterrows()]
        tail = [list(v) for _, v in _self.tail(half_of_max_rows_to_show).iterrows()]
        data = [*head, truncation_marker * _self.shape[1], *tail]
    else:
        data = [list(v) for _, v in _self.iterrows()]

    adjusted_data = adjust_table_width(
        data=data,
        headers=list(_self.columns),
        dtypes=[str(t) for t in _self.dtypes],
        max_col_width=max_col_width,
        max_total_width=console_width(max_total_width),
    )
    columns_not_shown = ""
    truncated_data = adjusted_data.truncated_data
    truncated_headers = adjusted_data.truncated_headers
    truncated_headers = [inverted_replacement_names.get(c, c) for c in truncated_headers]
    truncated_dtypes = adjusted_data.truncated_dtypes
    if len(truncated_headers) != len(_self.columns):
        num_not_shown = len(_self.columns) - len(truncated_headers)
        not_shown = [
            f'"{e}"'
            for e in _self.columns[
                adjusted_data.included_columns : adjusted_data.included_columns + max_column_names_to_show
            ]
        ]
        if num_not_shown > max_column_names_to_show:
            not_shown.append("...")
        columns_not_shown = f" ({num_not_shown} columns not shown: {', '.join(not_shown)})."
        truncated_data = [row + truncation_marker for row in truncated_data]
        truncated_headers += truncation_marker
        truncated_dtypes += truncation_marker
    headers_with_dtype = [f"{h}\n{d}" for h, d in zip(truncated_headers, truncated_dtypes, strict=True)]
    contains_duplicates_string = (
        f" (with {number_duplicated_indices} index duplicates)." if has_duplicated_index else "."
    )
    class_and_shape_info = (
        f"{self.__class__.__name__} with {self.shape[0]} rows, "
        f"{_self.shape[1] - number_index_levels} columns, and {number_index_levels} index columns{contains_duplicates_string}"
    )
    truncated_df = pd.DataFrame.from_records(truncated_data, columns=truncated_headers)
    headers_with_dtype.insert(number_index_levels, "|\n|")
    truncated_df.insert(number_index_levels, "|", "|")
    return f"{tabulate(truncated_df.to_numpy(), headers_with_dtype, showindex=False)}\n{class_and_shape_info}{columns_not_shown}"


def adjust_table_width(
    data: list[list[str]],
    headers: list[str],
    dtypes: list[str],
    max_total_width: int | None = None,
    max_col_width: int | None = None,
) -> AdjustedTableData:
    """Adjust table width to fit screen or given width."""
    # Truncate individual columns to max_col_width
    truncated_headers = truncate_data([headers], max_col_width)
    truncated_dtypes = truncate_data([dtypes], max_col_width)
    truncated_data = truncate_data(data, max_col_width)

    max_total_width = int(1e5) if max_total_width is None else max_total_width
    # Calculate the cumulative width of the columns after truncation
    cumulative_width = 0
    included_columns = 0
    for index, column in enumerate(zip(*(truncated_data + truncated_dtypes + truncated_headers), strict=True)):
        column_width = max(len(str(x)) + 4 for x in column)
        if (new_width := cumulative_width + column_width) <= max_total_width:
            cumulative_width = new_width
            included_columns = index + 1
        else:
            break

    # Slice the data and headers to include only the columns that fit into the max_total_width
    final_data = [row[:included_columns] for row in truncated_data]
    final_headers = truncated_headers[0][:included_columns]
    final_dtypes = truncated_dtypes[0][:included_columns]

    return AdjustedTableData(final_data, final_headers, final_dtypes, included_columns)


def truncate_data(data: list[list[str]], max_col_width: int | None) -> list[list[str]]:
    """Truncate data to max_col_width."""
    truncated_data = []
    for row in data:
        new_row = []
        for item in row:
            item_str = str(item)
            if max_col_width is not None:
                new_row.append(
                    item_str[: max_col_width - 3] + "..."
                    if len(item_str) > max_col_width
                    else item_str[:max_col_width],
                )
            else:
                new_row.append(item_str)
        truncated_data.append(new_row)
    return truncated_data


# placed at the beginning of every HTML representation
_html_header = """<div>
<style scoped>
    .dataframe tbody tr th:only-of-type {
        vertical-align: middle;
    }

    .dataframe tbody tr th {
        vertical-align: top;
    }

    .dataframe thead th {
        text-align: right;
    }
</style>"""

# placed at the end of every HTML representation
_html_footer = "</div>"


def tohtml(self: "pr.RangeFrame", **kwargs) -> str:
    """Return HTML representation, fit for Jupyter Notebook. Kwargs are passed to pandas.DataFrame.to_html."""
    number_index_levels = self.index.nlevels
    number_duplicated_indices = self.index.duplicated().sum()
    has_duplicated_index = number_duplicated_indices > 0

    html_max_rows = options.get_option("html_max_rows")
    if html_max_rows is None:
        html_max_rows = options.get_option("max_rows_to_show")
    html_max_cols = options.get_option("html_max_cols")
    max_column_names_to_show = options.get_option("max_column_names_to_show")

    kwargs.setdefault("max_rows", html_max_rows)
    kwargs.setdefault("max_cols", html_max_cols)
    h = pd.DataFrame.to_html(self, **kwargs)

    missing_message = ""
    if html_max_cols < len(self.columns):
        ## some columns are not shown in the HTML representation, let's find out which ones.
        # We could infer them from the DataFrame and html_max_cols, but let's interrogate the HTML representation
        # by Pandas. The current Pandas behavior is a bit strange (if you ask for 7 columns, it will show 6)
        # so determine it dynamically may be more robust for the future

        th_raw = re.findall(r"<thead>.*?</thead>", h, flags=re.DOTALL)[0]
        th_cells = re.findall(r"<th[^>]*>(.*?)</th>", th_raw, flags=re.DOTALL)

        # Decode entities like &amp; and strip whitespace
        header_vals = [_html.unescape(t).strip() for t in th_cells]
        header_vals = header_vals[1:]  # drop the index column
        displayed = {h for h in header_vals if h != "…"}  # “…” is the placeholder

        # Which DataFrame columns did NOT make it into the header?
        missing = [c for c in map(str, self.columns) if c not in displayed]

        missing_message = ""
        if missing:
            sample = ", ".join(f'"{c}"' for c in missing[:max_column_names_to_show])
            if len(missing) > max_column_names_to_show:
                sample += ", ..."
            missing_message = f" ({len(missing)} columns not shown: {sample})"

    contains_duplicates_string = (
        f" (with {number_duplicated_indices} index duplicates)." if has_duplicated_index else "."
    )

    info = (
        f"<p>"
        f"{self.__class__.__name__} with {self.shape[0]} rows, "
        f"{self.shape[1]} columns, and {number_index_levels} index columns{contains_duplicates_string}{missing_message}"
        f"</p>"
    )

    if reasons := InvalidRangesReason.formatted_reasons_list(self):
        info += f"<p>Invalid ranges:</p><p>{reasons}</p>"

    # now cowardly checking if self is a PyRanges class (rather than RangeFrame) without importing PyRanges object
    chrom_info: Callable[[], str] | None = getattr(self, "_chrom_and_strand_info", None)
    if chrom_info is not None:  # “callable” is implied by the annotation
        info += f"<p>{chrom_info()}.</p>"
        # adding something like:
        # Contains 6 chromosomes and 2 strands.
        # Contains 6 chromosomes and 3 strands (including non-genomic strands: asd).

    return _html_header + h + info + _html_footer



import itertools
from collections.abc import Iterable

import pandas as pd

import pyranges as pr  # noqa: F401

# thanks to Devon Ryan at https://bioinformatics.stackexchange.com/questions/3583/what-is-the-fastest-way-to-get-the-reverse-complement-of-a-dna-sequence-in-pytho
complement = str.maketrans("ACTGactg", "TGACtgac")
rnacomplement = str.maketrans("ACUGacug", "UGACugac")

# build alternative genetic code translation tables based on NCBI codes
GENETIC_CODES = {}  # will store dict of dicts: {genetic_code_id: {codon:aminoacid} }
GENETIC_CODE_AAS = {
    "1": "FFLLSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
    "2": "FFLLSSSSYY**CCWWLLLLPPPPHHQQRRRRIIMMTTTTNNKKSS**VVVVAAAADDEEGGGG",
    "3": "FFLLSSSSYY**CCWWTTTTPPPPHHQQRRRRIIMMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
    "4": "FFLLSSSSYY**CCWWLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
    "5": "FFLLSSSSYY**CCWWLLLLPPPPHHQQRRRRIIMMTTTTNNKKSSSSVVVVAAAADDEEGGGG",
    "6": "FFLLSSSSYYQQCC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
    "9": "FFLLSSSSYY**CCWWLLLLPPPPHHQQRRRRIIIMTTTTNNNKSSSSVVVVAAAADDEEGGGG",
    "10": "FFLLSSSSYY**CCCWLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
    "11": "FFLLSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
    "12": "FFLLSSSSYY**CC*WLLLSPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
    "13": "FFLLSSSSYY**CCWWLLLLPPPPHHQQRRRRIIMMTTTTNNKKSSGGVVVVAAAADDEEGGGG",
    "14": "FFLLSSSSYYY*CCWWLLLLPPPPHHQQRRRRIIIMTTTTNNNKSSSSVVVVAAAADDEEGGGG",
    "16": "FFLLSSSSYY*LCC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
    "21": "FFLLSSSSYY**CCWWLLLLPPPPHHQQRRRRIIMMTTTTNNNKSSSSVVVVAAAADDEEGGGG",
    "22": "FFLLSS*SYY*LCC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
    "23": "FF*LSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
    "24": "FFLLSSSSYY**CCWWLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSSKVVVVAAAADDEEGGGG",
    "25": "FFLLSSSSYY**CCGWLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
    "26": "FFLLSSSSYY**CC*WLLLAPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
    "27": "FFLLSSSSYYQQCCWWLLLAPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
    "28": "FFLLSSSSYYQQCCWWLLLAPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
    "29": "FFLLSSSSYYYYCC*WLLLAPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
    "30": "FFLLSSSSYYEECC*WLLLAPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
    "31": "FFLLSSSSYYEECCWWLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
    "33": "FFLLSSSSYYY*CCWWLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSSKVVVVAAAADDEEGGGG",
}

for gc_code in GENETIC_CODE_AAS:
    GENETIC_CODES[gc_code] = {"---": "-"}
    for codon_index, combo in enumerate(
        itertools.product("TCAG", repeat=3),
    ):  # iterates over ('T', 'T, 'T),  ('T', 'T', 'C'), ... etc
        codon = "".join(combo)
        GENETIC_CODES[gc_code][codon] = GENETIC_CODE_AAS[gc_code][codon_index]
    GENETIC_CODES[gc_code + "+U"] = GENETIC_CODES[gc_code].copy()
    GENETIC_CODES[gc_code + "+U"]["TGA"] = "U"


def reverse_complement(
    seq: str | pd.Series | Iterable[str],
    *,
    is_rna: bool = False,
    check: bool = False,
) -> str | pd.Series:
    """Reverse complement a DNA sequence.

    The input sequence is in DNA alphabet, upper or lowercase (ATGCatgc). The case is maintained in output.
    Other characters are left unchanged.

    Parameters
    ----------
    seq : str | Iterable[str]
        nucleotide sequence in DNA format (characters: ATGCatgc), or an iterable of such sequences, e.g. a Series

    is_rna : bool
        use this to provide the input seq in RNA format instead (characters: AUGCaugc)

    check : bool
        check if the input string contains only characters present in the translation table, raising an error if not

    Returns
    -------
    revcompseq : str
        Reverse complement nucleotide sequence in DNA format (or RNA if is_rna was set to True).
        If an iterable was provided, a Series of reverse complement sequences is returned.

    Examples
    --------
    >>> pr.seqs.reverse_complement("ATGAAATTTGGGTGA")
    'TCACCCAAATTTCAT'

    >>> pr.seqs.reverse_complement("AUGAAAUUUGGGUGA", is_rna=True)
    'UCACCCAAAUUUCAU'

    >>> pr.seqs.reverse_complement("aaaATCcccGGG")
    'CCCgggGATttt'

    >>> pr.seqs.reverse_complement("ATCWWWCCCTTT")
    'AAAGGGWWWGAT'

    >>> pr.seqs.reverse_complement("AUGAAATGGGTGA", check=True)
    Traceback (most recent call last):
    ...
    ValueError: One or more characters in the input string are not present in the translation table.

    >>> some_seqs=["ATGAAATTTGGGTGA", "AAAGAAATGGGTGACCCCC"]
    >>> pr.seqs.reverse_complement(some_seqs)
    0        TCACCCAAATTTCAT
    1    GGGGGTCACCCATTTCTTT
    dtype: object

    If a Series is provided, the output Series preserve its index:

    >>> some_seqs = pd.Series(["ATGAAATTTGGGTGA", "AAAGAAATGGGTGACCCCC"], index=["s1", "s2"])
    >>> pr.seqs.reverse_complement(some_seqs)
    s1        TCACCCAAATTTCAT
    s2    GGGGGTCACCCATTTCTTT
    dtype: object

    """

    def _rev_comp(seq: str, transtable: dict, *, check: bool) -> str:
        if check:
            accepted_chars = {chr(charint) for charint in transtable}
            if not set(seq).issubset(accepted_chars):
                msg = "One or more characters in the input string are not present in the translation table."
                raise ValueError(msg)
        return seq.translate(transtable)[::-1]

    def _vectorized_rev_comp(seqs: Iterable[str], transtable: dict, *, check: bool) -> pd.Series:
        seqs = pd.Series(list(seqs)) if not isinstance(seqs, pd.Series) else seqs
        if check:
            accepted_chars = {chr(charint) for charint in transtable}
            wrong_seq_sel = seqs.str.contains(f"[^{accepted_chars}]")
            if wrong_seq_sel.any():
                example = str(seqs[wrong_seq_sel].head(1)).split("\n")[0]
                msg = (
                    f"One or more characters in {wrong_seq_sel.sum()} input string(s) are not present in the "
                    f"translation table, such as: {example}"
                )
                raise ValueError(msg)

        return seqs.str.translate(transtable).str[::-1]

    transtable = complement if not is_rna else rnacomplement
    return (
        _rev_comp(seq, transtable, check=check)
        if isinstance(seq, str)
        else _vectorized_rev_comp(seq, transtable, check=check)
    )


GENETIC_CODE_OPTION = str | int | dict


def translate(  # noqa: C901
    seq: str | pd.Series | Iterable[str],
    *,
    genetic_code: GENETIC_CODE_OPTION = "1",
    unknown: str = "X",
    sanitize: bool = True,
    cache: int | bool = False,
) -> str | pd.Series:
    """Translate a coding sequence into protein.

    The input sequence is expected to be a DNA sequence in uppercase format (characters: ATGC).
    Incomplete codons at the end of the sequence, as well as non-canonical codons, result in the unknown character "X".

    Parameters
    ----------
    seq : str | Iterable[str]
        nucleotide sequence in uppercase DNA format (characters: ATGC), or an iterable of such sequences, e.g. a Series

    genetic_code : int | str | dict
        int or string-converted NCBI index for genetic code (https://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi)
        or dictionary with keys for each codon, values are amino acids (remember to
        include the translation for gaps ``'---':'-'``)
        or string-converted NCBI index with a '+U' suffix to have UGA as selenocysteine (U character)

    unknown : str | None
        codons that are not found in the genetic code table will be translated as this character
        if None, finding an unknown codon will raise an exception instead

    sanitize : bool
        Whether the input is converted to DNA uppercase (from lowercase and/or RNA) to ensure translation works.
        Set to False if you are sure sequences are already DNA uppercase, to speed up execution

    cache : bool | int
        speeds up translation by caching the result of translation of multicodon strings.
        The 1st time, the function is slow since precomputing all results; then, it is ~3 times faster than non-caching translate.
        With cache=True, all 3-codon translations are cached (memory 10Mb, precompute ~ 230ms).
        Provide an int to define how many codons to cache; this is approx the speedup that will be obtained.
        Note: memory and precomputing grow exponentially with the N of codons cached;
        You may use pyranges.seqs.clear_kmer_memory() to free memory.

    Returns
    -------
    pep : str
        Protein sequence resulting from translation, with gaps as '-' and unknown characters as 'X'.
        If an iterable was provided, a Series of protein sequences is returned.

    Examples
    --------
    >>> pr.seqs.translate("ATGAAATTTGGGTGA")
    'MKFG*'

    >>> pr.seqs.translate("ATGTTGCTGAA")
    'MLLX'

    Translate with the vertebrate mithochondrial genetic code:

    >>> pr.seqs.translate("ATGAAATTTGGGTGA", genetic_code=2)
    'MKFGW'

    Translate with a custom genetic code (all codons starting with A are translated as A, the rest as Q):

    >>> gc={codon:'A' if codon.startswith('A') else 'Q' for codon in map(''.join, itertools.product("TCAG", repeat=3))}
    >>> pr.seqs.translate("ATGAAATTTGGGTGA", genetic_code=gc)
    'AAQQQ'

    >>> pr.seqs.translate("AUGAAATTtGGGTGA", sanitize=False)
    'XKXG*'

    >>> pr.seqs.translate("AUGAAATTtGGGTGA", sanitize=False, unknown=None)
    Traceback (most recent call last):
    ...
    ValueError: translate ERROR cannot find codon AUG (pos 0-3) in genetic_code!

    Output for list of sequences is a Series:

    >>> pr.seqs.translate(['ACTGCATAA', 'ATGGGGTACTAG'])
    0     TA*
    1    MGY*
    dtype: object

    >>> pr.seqs.translate(['AAUUUtACTGCACTACGACTAGCTAC', 'ACACTGACTGACTATCTGATCGAC'], sanitize=True)
    0    NFTALRLAX
    1     TLTDYLID
    dtype: object

    When the input is a Series, the output is a Series with the same index:

    >>> x = pd.Series(['ACTGCATAA', 'ATGGGGTACTAG'], index=['s1', 's2'])
    >>> pr.seqs.translate(x)
    s1     TA*
    s2    MGY*
    dtype: object

    Caching makes sense when translating many sequences.
    For large enough data, translate with ``cache=True`` is 3x faster than non-caching translate:

    >>> import random
    >>> random.seed(42)
    >>> many_seqs = [ "".join(random.choices("ATGC", k=1000)) for _ in range(100000) ]
    >>> translated_seqs = pr.seqs.translate(many_seqs, cache=True)
    >>> translated_seqs[0][:50]
    'DRHKKEHVLTRLRGINVIGDRANYLS*CYESYRQGDS*PGRDTLPYHV*D'

    """

    def _translate_noncached(seq: str, *, sanitize: bool = False) -> str:
        seq = seq.upper().replace("U", "T") if sanitize else seq
        output = []
        for pos in range(0, len(seq), 3):
            codon = seq[pos : pos + 3]
            if codon in codon_table:
                output.append(codon_table[codon])
            elif unknown is not None:
                output.append(unknown)
            else:
                msg = f"translate ERROR cannot find codon {codon} (pos {pos}-{pos + 3}) in genetic_code!"
                raise ValueError(msg)
        return "".join(output)

    def _translate_cached(
        seq: str,
        *,
        sanitize: bool = False,
    ) -> str:
        seq = seq.upper().replace("U", "T") if sanitize else seq
        output = []
        for pos in range(0, len(seq), step):
            multicodon = seq[pos : pos + step]
            output.append(
                kmer_codon_table[multicodon]
                if multicodon in kmer_codon_table
                # for final bits of sequences which are shorter than k codons, and also for sequences containing Ns:
                else _translate_noncached(multicodon, sanitize=False),
            )
        return "".join(output)

    def _vectorized_translate(
        seqs: Iterable[str],
        *,
        sanitize: bool,
    ) -> pd.Series:
        seqs = pd.Series(list(seqs)) if not isinstance(seqs, pd.Series) else seqs
        seqs = seqs.str.upper().str.replace("U", "T") if sanitize else seqs

        return (
            seqs.apply(_translate_noncached)
            if not cache  # above: no cache
            else seqs.apply(_translate_cached)
        )

    if isinstance(genetic_code, dict):
        codon_table = genetic_code
    elif str(genetic_code) in GENETIC_CODES:
        codon_table = GENETIC_CODES[str(genetic_code)]
    else:
        msg = f"translate ERROR genetic_code input not recognized: {genetic_code}"
        raise ValueError(msg)

    if cache:
        # precomputing all possible translations for k codons, if not computed already (see _get_kmer_codon_table)
        cache = 3 if cache is True else cache
        kmer_codon_table = _get_kmer_codon_table(genetic_code, cache)
        step = cache * 3

    return (
        _translate_noncached(seq, sanitize=sanitize)
        if (isinstance(seq, str) and not cache)
        else (
            _translate_cached(seq, sanitize=sanitize)
            if isinstance(seq, str) and cache
            else _vectorized_translate(seq, sanitize=sanitize)
        )
    )


kmer_codon_tables = {}


def clear_kmer_memory() -> None:
    """Clear memory used by the translation cache."""
    kmer_codon_tables.clear()


def _get_kmer_codon_table(genetic_code, k) -> dict[str, str]:
    if (genetic_code, k) not in kmer_codon_tables:
        all_codons = list(GENETIC_CODES["1"].keys())
        codon_table_multik = {
            multicodon: translate(multicodon, genetic_code=genetic_code)
            for multicodon in map("".join, itertools.product(*(all_codons for i in range(k))))
        }
        kmer_codon_tables[(genetic_code, k)] = codon_table_multik

    return kmer_codon_tables[(genetic_code, k)]



from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import ruranges  # Rust extension

from pyranges.core.names import (
    CHROM_COL,
    END_COL,
    START_COL,
    STRAND_COL,
)
from pyranges.core.pyranges_helpers import ensure_pyranges, factorize_binary

if TYPE_CHECKING:
    from pyranges import PyRanges

cumsum_start = "_local_start"
cumsum_end = "_local_end"
out_local_start = "Start_local"
out_local_end = "End_local"
out_local_strand = "Strand_local"


def _map_to_global_pandas(
    local_gr: "PyRanges",
    global_gr: "PyRanges",
    global_on: str,
    *,
    local_on: str = "Chromosome",
    keep_id: bool = False,
    keep_loc: bool = False,
) -> "PyRanges":
    local_has_strand, global_has_strand = local_gr.has_strand, global_gr.has_strand
    original_order, original_columns = local_gr.index.copy(), local_gr.columns.copy()

    local_gr = local_gr.copy()
    local_gr["_temp_index_"] = local_gr.index.copy()

    global_gr = global_gr.get_with_loc_columns(global_on)

    m = local_gr.merge(
        global_gr,
        left_on=local_on,
        right_on=global_on,
        suffixes=("_local", ""),
    )

    z = m.slice_ranges(  # type: ignore[reportCallIssue]
        group_by="_temp_index_",
        start=m[out_local_start].to_numpy(),
        end=m[out_local_end].to_numpy(),
        use_strand=global_has_strand,
    )

    cols_to_drop = (
        [CHROM_COL + "_local"]
        + ([] if keep_id else [global_on])
        + ([] if keep_loc else [out_local_start, out_local_end])
    )

    if local_has_strand and global_has_strand:
        z[STRAND_COL] = np.where(z[out_local_strand] == z[STRAND_COL], "+", "-")
        if not keep_loc:
            cols_to_drop.append(out_local_strand)

    z = z.set_index("_temp_index_", drop=True)
    z.index.name = local_gr.index.name
    if (z := z.drop(columns=cols_to_drop)) is None:
        raise RuntimeError
    z = z.loc[original_order.intersection(z.index, sort=False)]

    bottom_cols: list[str] = []
    if keep_id:
        bottom_cols.append(global_on)
    elif (global_on + "_local") in z.columns:
        z = z.rename(columns={global_on + "_local": global_on})
    if keep_loc:
        bottom_cols += [out_local_start, out_local_end]
        if local_has_strand and global_has_strand:
            bottom_cols.append(out_local_strand)
    if global_has_strand and not local_has_strand:
        bottom_cols.append(STRAND_COL)

    cols_output = [c for c in original_columns if c not in bottom_cols] + bottom_cols
    return ensure_pyranges(z[cols_output])


# buggy:
def _map_to_global_ruranges(
    local_gr: "PyRanges",
    global_gr: "PyRanges",
    global_on: str,
    *,
    local_on: str = "Chromosome",
    keep_id: bool = False,
    keep_loc: bool = False,
) -> "PyRanges":
    """Lift intervals from local_gr to genomic coordinates using global_gr exon annotations.

    Lift intervals in *local_gr* (coordinates relative to a transcript) onto absolute genomic coordinates using the exon annotation in *global_gr*.
    """
    local_has_strand = local_gr.has_strand
    global_has_strand = global_gr.has_strand

    global_cum = global_gr.group_cumsum(
        group_by=global_on,
        use_strand="auto",
        cumsum_start_column=cumsum_start,
        cumsum_end_column=cumsum_end,
        keep_order=True,  # False,
    )
    ex_df = global_cum.copy()

    ex_df["__global_start__"] = ex_df[START_COL]
    ex_df["__global_end__"] = ex_df[END_COL]
    ex_df[START_COL] = ex_df[cumsum_start]
    ex_df[END_COL] = ex_df[cumsum_end]

    tmp_tx_id = "__txid__"
    local_df = local_gr.copy()

    ex_df[tmp_tx_id] = ex_df[global_on]
    local_df[tmp_tx_id] = local_df[local_on]

    q_tx_code, ex_tx_code = factorize_binary(local_df, ex_df, by=tmp_tx_id)

    ex_chr_code, chr_table = pd.factorize(ex_df[CHROM_COL], sort=False)
    ex_chr_code = ex_chr_code.astype(ex_tx_code.dtype, copy=False)

    ex_local_start = ex_df[START_COL].to_numpy(copy=False)
    ex_local_end = ex_df[END_COL].to_numpy(copy=False)
    ex_genome_start = ex_df["__global_start__"].to_numpy(copy=False)
    ex_genome_end = ex_df["__global_end__"].to_numpy(copy=False)
    ex_fwd = (
        (ex_df[STRAND_COL] == "+").to_numpy(np.bool_, copy=False)
        if global_has_strand
        else np.ones(len(ex_df), dtype=bool)
    )

    q_start = local_df[START_COL].to_numpy(copy=False)
    q_end = local_df[END_COL].to_numpy(copy=False)
    q_fwd = (
        (local_df[STRAND_COL] == "+").to_numpy(np.bool_, copy=False)
        if local_has_strand
        else np.ones(len(local_df), dtype=bool)
    )

    keep_idx, out_start, out_end, out_strand_bool = ruranges.map_to_global(
        groups=ex_tx_code,
        starts=ex_local_start,
        ends=ex_local_end,
        strand=ex_fwd,
        groups2=q_tx_code,
        starts2=q_start,
        ends2=q_end,
        chr_code2=ex_chr_code,
        genome_start2=ex_genome_start,
        genome_end2=ex_genome_end,
        strand2=q_fwd,
    )

    if len(keep_idx) == 0:
        return ensure_pyranges(local_df.head(0))

    max_code = int(ex_tx_code.max()) + 1
    tx2chr_code = np.empty(max_code, dtype=ex_chr_code.dtype)
    tx2chr_code[ex_tx_code] = ex_chr_code
    out_chr_code = tx2chr_code[q_tx_code[keep_idx]]
    out_chr_str = chr_table[out_chr_code]

    mapped_df = local_gr.take(keep_idx).copy()  # type: ignore[arg-type]

    if keep_id:
        mapped_df[global_on] = mapped_df[CHROM_COL].copy()
    if keep_loc:
        mapped_df[out_local_start] = mapped_df[START_COL].copy()
        mapped_df[out_local_end] = mapped_df[END_COL].copy()
        if local_has_strand and global_has_strand:
            mapped_df[out_local_strand] = mapped_df[STRAND_COL].copy()

    mapped_df[CHROM_COL] = out_chr_str
    mapped_df[START_COL] = out_start
    mapped_df[END_COL] = out_end

    if local_has_strand or global_has_strand:
        mapped_df[STRAND_COL] = np.where(out_strand_bool, "+", "-")

    if tmp_tx_id in mapped_df.columns:
        mapped_df = mapped_df.drop(columns=[tmp_tx_id])
        if mapped_df is None:
            raise ValueError

    return ensure_pyranges(mapped_df)



"""Module of example data.

See Also
--------
pyranges.random : generate random PyRanges

Examples
--------
>>> pr.example_data
Available example data:
-----------------------
example_data.chipseq            : Example ChIP-seq data.
example_data.chipseq_background : Example ChIP-seq data.
example_data.chromsizes         : Example chromsizes data (hg19).
example_data.ensembl_gtf        : Example gtf file from Ensembl.
example_data.interpro_hits      : Example of InterPro protein hits.
example_data.rfam_hits          : Example of RNA motifs (Rfam) as 1-based dataframe.
example_data.f1                 : Example bed file.
example_data.f2                 : Example bed file.
example_data.aorta              : Example ChIP-seq data.
example_data.aorta2             : Example ChIP-seq data.
example_data.ncbi_gff           : Example NCBI GFF data.
example_data.ncbi_fasta         : Example NCBI fasta.
example_data.files              : Return a dict of basenames to file paths of available data.

>>> pr.example_data.f1
  index  |    Chromosome      Start      End  Name         Score  Strand
  int64  |    category        int64    int64  object       int64  category
-------  ---  ------------  -------  -------  ---------  -------  ----------
      0  |    chr1                3        6  interval1        0  +
      1  |    chr1                5        7  interval2        0  -
      2  |    chr1                8        9  interval3        0  +
PyRanges with 3 rows, 6 columns, and 1 index columns.
Contains 1 chromosomes and 2 strands.


"""

import logging
import sys
import tempfile
import typing
from importlib.resources import files
from pathlib import Path
from typing import Any, ClassVar

import pandas as pd

import pyranges as pr
from pyranges.core.pyranges_helpers import ensure_pyranges

if typing.TYPE_CHECKING:
    import pyfaidx

    from pyranges.core.pyranges_main import PyRanges

logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


# Define a custom class property decorator
class ClassProperty:
    def __init__(self, fget: Any) -> None:
        self.fget = fget
        # Propagate the docstring from the getter function
        self.__doc__ = fget.__doc__

    def __get__(self, instance: Any | None, owner: type[Any]) -> Any:
        return self.fget(owner)


class ExampleData:
    _files: ClassVar[dict[str, Path]] = {}

    def __repr__(self) -> str:
        methods_info = "Available example data:\n-----------------------\n"
        items = [
            (name, prop)
            for name, prop in self.__class__.__dict__.items()
            if isinstance(prop, property) and not name.startswith("_")
        ]
        max_name_len = max(len(name) for name, _ in (*items, ("files", "")))
        for name, prop in items:
            doc_line = prop.fget.__doc__.split("\n")[0] if prop.fget.__doc__ else ""
            methods_info += f"example_data.{name:<{max_name_len}} : {doc_line}\n"
        for name in ["files"]:
            doc_line = self.__class__.__dict__[name].__doc__.strip().split("\n")[0]
            methods_info += f"example_data.{name:<{max_name_len}} : {doc_line}\n"
        return methods_info.strip()

    @ClassProperty
    def files(self) -> dict[str, Path]:
        """Return a dict of basenames to file paths of available data.

        Examples
        --------
        >>> bam = ExampleData.files["smaller.bam"]
        >>> bam.exists()
        True

        """
        if ExampleData._files:
            return ExampleData._files

        paths = []
        for f in files("pyranges").joinpath("data").iterdir():
            if not isinstance(f, Path):
                msg = f"Expected Path, got {type(f)}"
                raise TypeError(msg)
            if "__" not in f.name:
                paths.append(f)
        ExampleData._files = {f.name: Path(f) for f in paths}
        return ExampleData._files

        paths = []
        for f in files("pyranges").joinpath("data").iterdir():
            if not isinstance(f, Path):
                msg = f"Expected Path, got {type(f)}"
                raise TypeError(msg)
            if "__" not in f.name:
                paths.append(f)
        ExampleData._files = {f.name: Path(f) for f in paths}
        return ExampleData._files

    @staticmethod
    def _read_bed_from_string(contents: str) -> "PyRanges":
        with tempfile.NamedTemporaryFile("w", encoding="utf-8") as f:
            f.write(contents)
            f.flush()
            return pr.read_bed(Path(f.name))

    @staticmethod
    def _read_gtf_from_string(contents: str) -> "PyRanges":
        with tempfile.NamedTemporaryFile("w", encoding="utf-8") as f:
            f.write(contents)
            f.flush()
            return pr.read_gtf(Path(f.name))

    @staticmethod
    def _read_tsv_from_string(contents: str) -> "pd.DataFrame":
        with tempfile.NamedTemporaryFile("w", encoding="utf-8") as f:
            f.write(contents)
            f.flush()
            return pd.read_csv(Path(f.name), sep="\t")

    @property
    def chipseq(self) -> "pr.PyRanges":
        """Example ChIP-seq data.

        From the SICER software.
        """
        contents = """chr8	28510032	28510057	U0	0	-
chr7	107153363	107153388	U0	0	-
chr5	135821802	135821827	U0	0	-
chr14	19418999	19419024	U0	0	-
chr12	106679761	106679786	U0	0	-
chr21	40099618	40099643	U0	0	+
chr8	22714402	22714427	U0	0	-
chr19	19571102	19571127	U0	0	+
chr3	140986358	140986383	U0	0	-
chr10	35419784	35419809	U0	0	-
chr4	98488749	98488774	U0	0	+
chr11	22225193	22225218	U0	0	+
chr1	38457520	38457545	U0	0	+
chr1	80668132	80668157	U0	0	-
chr2	152562484	152562509	U0	0	-
chr4	153155301	153155326	U0	0	+
chr9	120803448	120803473	U0	0	+
chr6	89296757	89296782	U0	0	-
chr1	194245558	194245583	U0	0	+
chr8	57916061	57916086	U0	0	+"""
        return self._read_bed_from_string(contents)

    @property
    def chipseq_background(self) -> "pr.PyRanges":
        """Example ChIP-seq data.

        From the SICER software.
        """
        contents = """chr7	20246668	20246693	U0	0	+
chr1	39036822	39036847	U0	0	+
chr19	47109000	47109025	U0	0	-
chr10	90059861	90059886	U0	0	-
chr3	55648137	55648162	U0	0	+
chr7	91135110	91135135	U0	0	+
chr13	100938475	100938500	U0	0	+
chr3	115816130	115816155	U0	0	+
chr19	43528773	43528798	U0	0	+
chr10	73781101	73781126	U0	0	+"""
        return self._read_bed_from_string(contents)

    @property
    def chromsizes(self) -> "pr.PyRanges":
        """Example chromsizes data (hg19)."""
        contents = """chr1	0	249250621
chr2	0	243199373
chr3	0	198022430
chr4	0	191154276
chr5	0	180915260
chr6	0	171115067
chr7	0	159138663
chrX	0	155270560
chr8	0	146364022
chr9	0	141213431
chr10	0	135534747
chr11	0	135006516
chr12	0	133851895
chr13	0	115169878
chr14	0	107349540
chr15	0	102531392
chr16	0	90354753
chr17	0	81195210
chr18	0	78077248
chr20	0	63025520
chrY	0	59373566
chr19	0	59128983
chr22	0	51304566
chr21	0	48129895
chrM	0	16571"""
        return self._read_bed_from_string(contents)

    @property
    def ensembl_gtf(self) -> "pr.PyRanges":
        """Example gtf file from Ensembl."""
        contents = """#!genome-build GRCh38.p10
#!genome-version GRCh38
#!genome-date 2013-12
#!genome-build-accession NCBI:GCA_000001405.25
#!genebuild-last-updated 2017-06
1	havana	gene	11869	14409	.	+	.	gene_id "ENSG00000223972"; gene_version "5"; gene_name "DDX11L1"; gene_source "havana"; gene_biotype "transcribed_unprocessed_pseudogene";
1	havana	transcript	11869	14409	.	+	.	gene_id "ENSG00000223972"; gene_version "5"; transcript_id "ENST00000456328"; transcript_version "2"; gene_name "DDX11L1"; gene_source "havana"; gene_biotype "transcribed_unprocessed_pseudogene"; transcript_name "DDX11L1-202"; transcript_source "havana"; transcript_biotype "processed_transcript"; tag "basic"; transcript_support_level "1";
1	havana	exon	11869	12227	.	+	.	gene_id "ENSG00000223972"; gene_version "5"; transcript_id "ENST00000456328"; transcript_version "2"; exon_number "1"; gene_name "DDX11L1"; gene_source "havana"; gene_biotype "transcribed_unprocessed_pseudogene"; transcript_name "DDX11L1-202"; transcript_source "havana"; transcript_biotype "processed_transcript"; exon_id "ENSE00002234944"; exon_version "1"; tag "basic"; transcript_support_level "1";
1	havana	exon	12613	12721	.	+	.	gene_id "ENSG00000223972"; gene_version "5"; transcript_id "ENST00000456328"; transcript_version "2"; exon_number "2"; gene_name "DDX11L1"; gene_source "havana"; gene_biotype "transcribed_unprocessed_pseudogene"; transcript_name "DDX11L1-202"; transcript_source "havana"; transcript_biotype "processed_transcript"; exon_id "ENSE00003582793"; exon_version "1"; tag "basic"; transcript_support_level "1";
1	havana	exon	13221	14409	.	+	.	gene_id "ENSG00000223972"; gene_version "5"; transcript_id "ENST00000456328"; transcript_version "2"; exon_number "3"; gene_name "DDX11L1"; gene_source "havana"; gene_biotype "transcribed_unprocessed_pseudogene"; transcript_name "DDX11L1-202"; transcript_source "havana"; transcript_biotype "processed_transcript"; exon_id "ENSE00002312635"; exon_version "1"; tag "basic"; transcript_support_level "1";
1	havana	exon	112700	112804	.	-	.	gene_id "ENSG00000238009"; gene_version "6"; transcript_id "ENST00000471248"; transcript_version "1"; exon_number "2"; gene_name "AL627309.1"; gene_source "ensembl_havana"; gene_biotype "lincRNA"; transcript_name "AL627309.1-203"; transcript_source "havana"; transcript_biotype "lincRNA"; exon_id "ENSE00001957285"; exon_version "1"; transcript_support_level "5";
1	havana	exon	110953	111357	.	-	.	gene_id "ENSG00000238009"; gene_version "6"; transcript_id "ENST00000471248"; transcript_version "1"; exon_number "3"; gene_name "AL627309.1"; gene_source "ensembl_havana"; gene_biotype "lincRNA"; transcript_name "AL627309.1-203"; transcript_source "havana"; transcript_biotype "lincRNA"; exon_id "ENSE00001879696"; exon_version "1"; transcript_support_level "5";
1	ensembl	transcript	120725	133723	.	-	.	gene_id "ENSG00000238009"; gene_version "6"; transcript_id "ENST00000610542"; transcript_version "1"; gene_name "AL627309.1"; gene_source "ensembl_havana"; gene_biotype "lincRNA"; transcript_name "AL627309.1-205"; transcript_source "ensembl"; transcript_biotype "lincRNA"; tag "basic"; transcript_support_level "5";
1	ensembl	exon	133374	133723	.	-	.	gene_id "ENSG00000238009"; gene_version "6"; transcript_id "ENST00000610542"; transcript_version "1"; exon_number "1"; gene_name "AL627309.1"; gene_source "ensembl_havana"; gene_biotype "lincRNA"; transcript_name "AL627309.1-205"; transcript_source "ensembl"; transcript_biotype "lincRNA"; exon_id "ENSE00003748456"; exon_version "1"; tag "basic"; transcript_support_level "5";
1	ensembl	exon	129055	129223	.	-	.	gene_id "ENSG00000238009"; gene_version "6"; transcript_id "ENST00000610542"; transcript_version "1"; exon_number "2"; gene_name "AL627309.1"; gene_source "ensembl_havana"; gene_biotype "lincRNA"; transcript_name "AL627309.1-205"; transcript_source "ensembl"; transcript_biotype "lincRNA"; exon_id "ENSE00003734824"; exon_version "1"; tag "basic"; transcript_support_level "5";
1	ensembl	exon	120874	120932	.	-	.	gene_id "ENSG00000238009"; gene_version "6"; transcript_id "ENST00000610542"; transcript_version "1"; exon_number "3"; gene_name "AL627309.1"; gene_source "ensembl_havana"; gene_biotype "lincRNA"; transcript_name "AL627309.1-205"; transcript_source "ensembl"; transcript_biotype "lincRNA"; exon_id "ENSE00003740919"; exon_version "1"; tag "basic"; transcript_support_level "5";"""
        return self._read_gtf_from_string(contents)

    @property
    def interpro_hits(self) -> "pr.PyRanges":
        """Example of InterPro protein hits."""
        contents = """Chromosome	predictor	feature	Start	End
CAD5126873.1	SignalP_EUK	SignalP-noTM	1	18
CAD5126492.1	SignalP_EUK	SignalP-noTM	1	16
CAD5126498.1	SignalP_EUK	SignalP-noTM	1	21
CAD5126499.1	SignalP_EUK	SignalP-noTM	1	21"""
        d = pr.PyRanges(self._read_tsv_from_string(contents))
        d["Start"] -= 1
        return ensure_pyranges(d)

    @property
    def rfam_hits(self) -> "pd.DataFrame":
        """Example of RNA motifs (Rfam) as 1-based dataframe."""
        contents = """target_name	accession	query_name	accession.1	mdl	mdl_from	mdl_to	seq_from	seq_to	strand	trunc	pass	gc	bias	score	E-value	inc	description_of_target
rna-DGYR_LOCUS12552	-	GAIT	RF00179	cm	1	71	267	311	+	no	1	0.31	0.0	19.2	0.0067	!	-
rna-DGYR_LOCUS12552-2	-	GAIT	RF00179	cm	1	71	288	332	+	no	1	0.31	0.0	19.2	0.0067	!	-
rna-DGYR_LOCUS13738	-	REN-SRE	RF00180	cm	1	37	1641	1678	+	no	1	0.37	0.0	16.6	0.0015	!	-
rna-DGYR_LOCUS14091	-	IFN_gamma	RF00259	cm	1	169	137	43	-	no	1	0.27	0.7	19.6	0.004	!	-
rna-DGYR_LOCUS14091	-	snoZ30	RF00288	cm	1	97	547	616	+	no	1	0.33	0.0	18.5	0.005	!	-
rna-DGYR_LOCUS13734	-	snoR21	RF00352	cm	1	82	3276	3189	-	no	1	0.35	0.0	30.4	0.00027	!	-
rna-DGYR_LOCUS13734	-	HIV_GSL3	RF00376	cm	1	84	3356	3278	-	no	1	0.34	0.0	21.2	0.0022	!	-
rna-DGYR_LOCUS14091	-	SNORA53	RF00563	cm	1	248	833	748	-	no	1	0.29	0.1	15.4	0.0026	!	-
rna-DGYR_LOCUS13737	-	mir-320	RF00736	cm	1	76	1264	1310	+	no	1	0.28	0.1	15.6	0.0095	!	-
rna-DGYR_LOCUS13739	-	sR43	RF01128	hmm	7	38	432	468	+	-	6	0.43	0.0	10.7	0.0074	!	-
rna-DGYR_LOCUS13734	-	SNORD107	RF01164	cm	1	74	981	1052	+	no	1	0.25	0.6	21.2	0.0019	!	-
rna-DGYR_LOCUS14092	-	snR77	RF01181	hmm	17	61	572	617	+	-	6	0.33	0.7	12.9	0.0016	!	-
rna-DGYR_LOCUS14092	-	snR50	RF01190	cm	1	89	716	792	+	no	1	0.30	0.1	21.0	0.0085	!	-
rna-DGYR_LOCUS13737	-	snoR31	RF01288	cm	1	93	244	172	-	no	1	0.23	1.0	21.6	0.0045	!	-
rna-DGYR_LOCUS13738	-	CRISPR-DR42	RF01351	hmm	7	27	312	332	+	-	6	0.24	0.3	13.8	0.0017	!	-
rna-DGYR_LOCUS13734	-	rli62	RF01486	cm	1	135	276	401	+	no	1	0.36	0.0	20.1	0.0024	!	-
rna-DGYR_LOCUS13738	-	6S-Flavo	RF01685	cm	1	108	2414	2503	+	no	1	0.28	0.4	25.2	0.002	!	-
rna-DGYR_LOCUS13734	-	alpha_tmRNA	RF01849	cm	40	242	2457	2581	+	no	1	0.33	0.2	13.5	0.00042	!	-
rna-DGYR_LOCUS13734	-	STnc420	RF02054	cm	1	58	94	48	-	no	1	0.19	0.6	18.2	0.0076	!	-
rna-DGYR_LOCUS14095	-	MESTIT1_1	RF02148	hmm	60	124	174	112	-	-	6	0.25	1.1	11.7	0.0022	!	-
rna-DGYR_LOCUS14095-2	-	MESTIT1_1	RF02148	hmm	60	124	174	112	-	-	6	0.25	1.1	11.7	0.0022	!	-
rna-DGYR_LOCUS13738	-	PVT1_5	RF02168	hmm	17	117	739	635	-	-	6	0.26	0.1	13.6	0.0008	!	-
rna-DGYR_LOCUS13736	-	TtnuHACA18	RF02325	cm	1	130	376	263	-	no	1	0.36	0.0	28.3	0.00024	!	-
rna-DGYR_LOCUS13740	-	TtnuHACA18	RF02325	cm	1	130	538	425	-	no	1	0.36	0.0	28.3	0.00024	!	-
rna-DGYR_LOCUS13741	-	TtnuHACA18	RF02325	cm	1	130	538	425	-	no	1	0.36	0.0	28.3	0.00024	!	-
rna-DGYR_LOCUS12552	-	psRNA6	RF02350	cm	242	333	884	798	-	5'	2	0.21	8.4	11.5	0.0063	!	-
rna-DGYR_LOCUS12552-2	-	psRNA6	RF02350	cm	242	333	905	819	-	5'	2	0.21	8.4	11.5	0.0063	!	-
rna-DGYR_LOCUS13736	-	ToxI	RF02519	cm	1	34	393	436	+	no	1	0.32	0.0	19.4	0.005	!	-
rna-DGYR_LOCUS14092	-	OppA_thermometer	RF02777	cm	1	65	850	913	+	3'	3	0.25	0.2	12.7	0.0042	!	-
rna-DGYR_LOCUS13737	-	RT-8	RF03029	cm	1	92	163	77	-	no	1	0.33	0.0	21.4	0.0026	!	-
rna-DGYR_LOCUS13734	-	mir-5697	RF03269	cm	1	59	976	1026	+	no	1	0.18	3.6	18.7	0.005	!	-
rna-DGYR_LOCUS13737	-	mir-3047	RF03339	cm	1	61	550	600	+	no	1	0.33	0.0	18.2	0.0074	!	-
rna-DGYR_LOCUS13737	-	mir-3156	RF03360	cm	1	77	544	605	+	no	1	0.32	0.0	26.3	4.9e-05	!	-
rna-DGYR_LOCUS13737	-	MIR1523	RF04113	cm	1	92	530	612	+	no	1	0.30	0.1	18.8	0.0053	!	-
rna-DGYR_LOCUS13737	-	MIR8001	RF04255	cm	1	67	550	600	+	no	1	0.33	0.0	18.7	0.0068	!	-"""
        return self._read_tsv_from_string(contents)

    @property
    def f1(self) -> "pr.PyRanges":
        """Example bed file."""
        contents = """chr1	3	6	interval1	0	+
chr1	5	7	interval2	0	-
chr1	8	9	interval3	0	+"""
        return self._read_bed_from_string(contents)

    @property
    def f2(self) -> "pr.PyRanges":
        """Example bed file."""
        contents = """chr1	1	2	a	0	+
chr1	6	7	b	0	-"""
        with tempfile.NamedTemporaryFile("w", encoding="utf-8") as f:
            f.write(contents)
            f.flush()
            return pr.read_bed(Path(f.name))

    @property
    def aorta(self) -> "pr.PyRanges":
        """Example ChIP-seq data.

        From the epigenomics roadmap.
        """
        return pr.read_bed(ExampleData.files["aorta.bed"])  # type: ignore[index, misc]

    @property
    def aorta2(self) -> "pr.PyRanges":
        """Example ChIP-seq data.

        From the epigenomics roadmap.
        """
        return pr.read_bed(ExampleData.files["aorta2.bed"])  # type: ignore[index, misc]

    @property
    def ncbi_gff(self) -> "pr.PyRanges":
        """Example NCBI GFF data.

        Subset of the NCBI annotation of D.gyrociliatus assembly GCA_904063045.1.
        """
        return pr.read_gff3(ExampleData.files["ncbi.gff.gz"])  # type: ignore[index, misc]

    @property
    def ncbi_fasta(self) -> "pyfaidx.Fasta":
        """Example NCBI fasta.

        Subset of the NCBI D.gyrociliatus assembly GCA_904063045.1.

        A pyfaidx.Fasta object is returned. If not installed, an exception is raised.
        To retrieve the location of the fasta file, use pyranges.example_data.files['ncbi.fasta']
        """
        try:
            import pyfaidx  # type: ignore[import]
        except ImportError:
            LOGGER.exception(
                "To use this method, pyfaidx must be installed. To get just the fasta file path, use pyranges.example_data.files['ncbi.fasta']. Use `conda install -c bioconda pyfaidx` or `pip install pyfaidx` to install pyfaidx.",
            )
            sys.exit(1)

        # note: example data include ncbi.fasta.fai, the pyfaidx index
        return pyfaidx.Fasta(ExampleData.files["ncbi.fasta"])  # type: ignore[index, misc]


example_data = ExampleData()



from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import pandas as pd
import pandas.core.groupby  # type: ignore[name-defined]

from pyranges.core.names import return_pyranges_if_possible

if TYPE_CHECKING:
    import pyranges as pr


class PyRangesDataFrameGroupBy(pandas.core.groupby.DataFrameGroupBy):
    def __init__(self, pandas_groupby) -> None:
        self._pandas_groupby = pandas_groupby

    @property
    def pandas_groupby(self) -> pd.core.groupby.DataFrameGroupBy:
        """Return the underlying pandas groupby object."""
        return self.__dict__["_pandas_groupby"]

    def __getattr__(self, item) -> Any:
        # Handle attribute access, e.g., g.some_method()
        pd_grpby = self.__dict__["_pandas_groupby"]

        if hasattr(pd_grpby, item):
            attr = getattr(pd_grpby, item)
            if callable(attr):

                def wrapper(*args, **kwargs) -> Callable:
                    result = attr(*args, **kwargs)
                    return return_pyranges_if_possible(result)

                return wrapper
            return attr
        return None

    def __getitem__(self, key) -> "PyRangesDataFrameGroupBy":
        # Handle item access, e.g., g['column_name']
        result = self.pandas_groupby[key]
        return PyRangesDataFrameGroupBy(result)

    @return_pyranges_if_possible
    def agg(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.agg(*args, **kwargs)

    @return_pyranges_if_possible
    def aggregate(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.aggregate(*args, **kwargs)

    @return_pyranges_if_possible
    def all(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.all(*args, **kwargs)

    @return_pyranges_if_possible
    def any(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.any(*args, **kwargs)

    @return_pyranges_if_possible
    def apply(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        kwargs.setdefault("include_groups", False)
        return self.pandas_groupby.apply(*args, **kwargs)

    @return_pyranges_if_possible
    def bfill(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.bfill(*args, **kwargs)

    @return_pyranges_if_possible
    def cumcount(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.cumcount(*args, **kwargs)

    @return_pyranges_if_possible
    def cummax(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.cummax(*args, **kwargs)

    @return_pyranges_if_possible
    def cummin(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.cummin(*args, **kwargs)

    @return_pyranges_if_possible
    def cumprod(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.cumprod(*args, **kwargs)

    @return_pyranges_if_possible
    def cumsum(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.cumsum(*args, **kwargs)

    @return_pyranges_if_possible
    def describe(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.describe(*args, **kwargs)

    @return_pyranges_if_possible
    def diff(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.diff(*args, **kwargs)

    @return_pyranges_if_possible
    def ewm(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.ewm(*args, **kwargs)

    @return_pyranges_if_possible
    def expanding(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.expanding(*args, **kwargs)

    @return_pyranges_if_possible
    def ffill(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.ffill(*args, **kwargs)

    @return_pyranges_if_possible
    def fillna(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.fillna(*args, **kwargs)

    @return_pyranges_if_possible
    def filter(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.filter(*args, **kwargs)

    @return_pyranges_if_possible
    def first(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.first(*args, **kwargs)

    @return_pyranges_if_possible
    def get_group(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.get_group(*args, **kwargs)

    @return_pyranges_if_possible
    def head(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.head(*args, **kwargs)

    @return_pyranges_if_possible
    def idxmax(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.idxmax(*args, **kwargs)

    @return_pyranges_if_possible
    def idxmin(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.idxmin(*args, **kwargs)

    @return_pyranges_if_possible
    def last(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.last(*args, **kwargs)

    @return_pyranges_if_possible
    def max(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.max(*args, **kwargs)

    @return_pyranges_if_possible
    def mean(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.mean(*args, **kwargs)

    @return_pyranges_if_possible
    def median(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.median(*args, **kwargs)

    @return_pyranges_if_possible
    def min(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.min(*args, **kwargs)

    @return_pyranges_if_possible
    def ngroup(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.ngroup(*args, **kwargs)

    @return_pyranges_if_possible
    def nunique(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.nunique(*args, **kwargs)

    @return_pyranges_if_possible
    def ohlc(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.ohlc(*args, **kwargs)

    @return_pyranges_if_possible
    def pct_change(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.pct_change(*args, **kwargs)

    @return_pyranges_if_possible
    def pipe(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.pipe(*args, **kwargs)

    @return_pyranges_if_possible
    def prod(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.prod(*args, **kwargs)

    @return_pyranges_if_possible
    def quantile(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.quantile(*args, **kwargs)

    @return_pyranges_if_possible
    def rank(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.rank(*args, **kwargs)

    @return_pyranges_if_possible
    def resample(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.resample(*args, **kwargs)

    @return_pyranges_if_possible
    def rolling(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.rolling(*args, **kwargs)

    @return_pyranges_if_possible
    def sample(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.sample(*args, **kwargs)

    @return_pyranges_if_possible
    def sem(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.sem(*args, **kwargs)

    @return_pyranges_if_possible
    def shift(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.shift(*args, **kwargs)

    @return_pyranges_if_possible
    def size(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.size(*args, **kwargs)

    @return_pyranges_if_possible
    def skew(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.skew(*args, **kwargs)

    @return_pyranges_if_possible
    def std(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.std(*args, **kwargs)

    @return_pyranges_if_possible
    def sum(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.sum(*args, **kwargs)

    @return_pyranges_if_possible
    def tail(self, *args, **kwargs) -> "pr.PyRanges | pd.DataFrame | pd.Series":  # noqa: D102
        return self.pandas_groupby.tail(*args, **kwargs)



from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from pyranges.core.names import CHROM_COL, END_COL, START_COL
from pyranges.core.pyranges_helpers import ensure_pyranges

if TYPE_CHECKING:
    from pyranges import PyRanges


def tile_genome(
    chromsizes: "PyRanges | pd.DataFrame | dict[str | int, int]",
    tile_size: int,
    *,
    full_last_tile: bool = False,
) -> "PyRanges":
    """Create a tiled genome.

    Split the genome into adjacent non-overlapping tiles of a given size.

    Parameters
    ----------
    chromsizes : dict or PyRanges
        Dict or PyRanges describing the lengths of the chromosomes.

    tile_size : int
        Length of the tiles.

    full_last_tile : bool, default False
        Do not truncate the last tile to the end of the chromosome. Use to ensure size is consistent for all tiles.

    See Also
    --------
    pyranges.PyRanges.tile_ranges : split intervals into adjacent non-overlapping tiles.

    Examples
    --------
    >>> import pyranges as pr
    >>> chromsizes = pr.example_data.chromsizes
    >>> chromsizes
    index    |    Chromosome    Start    End
    int64    |    category      int64    int64
    -------  ---  ------------  -------  ---------
    0        |    chr1          0        249250621
    1        |    chr2          0        243199373
    2        |    chr3          0        198022430
    3        |    chr4          0        191154276
    ...      |    ...           ...      ...
    21       |    chr19         0        59128983
    22       |    chr22         0        51304566
    23       |    chr21         0        48129895
    24       |    chrM          0        16571
    PyRanges with 25 rows, 3 columns, and 1 index columns.
    Contains 25 chromosomes.

    >>> pr.tile_genome(chromsizes,int(1e6))
    index    |    Chromosome    Start     End
    int64    |    category      int64     int64
    -------  ---  ------------  --------  --------
    0        |    chr1          0         1000000
    1        |    chr1          1000000   2000000
    2        |    chr1          2000000   3000000
    3        |    chr1          3000000   4000000
    ...      |    ...           ...       ...
    3110     |    chrY          56000000  57000000
    3111     |    chrY          57000000  58000000
    3112     |    chrY          58000000  59000000
    3113     |    chrY          59000000  59373566
    PyRanges with 3114 rows, 3 columns, and 1 index columns.
    Contains 25 chromosomes.

    >>> pr.tile_genome(chromsizes,int(1e6), full_last_tile=True)
    index    |    Chromosome    Start     End
    int64    |    category      int64     int64
    -------  ---  ------------  --------  --------
    0        |    chr1          0         1000000
    1        |    chr1          1000000   2000000
    2        |    chr1          2000000   3000000
    3        |    chr1          3000000   4000000
    ...      |    ...           ...       ...
    3110     |    chrY          56000000  57000000
    3111     |    chrY          57000000  58000000
    3112     |    chrY          58000000  59000000
    3113     |    chrY          59000000  60000000
    PyRanges with 3114 rows, 3 columns, and 1 index columns.
    Contains 25 chromosomes.

    """
    # obtain a PyRanges with regular tiles
    if isinstance(chromsizes, dict):
        chromsize_dict = chromsizes
        df = pd.DataFrame({CHROM_COL: list(chromsizes), START_COL: 0, END_COL: list(chromsizes.values())})
        chromsizes = df
    else:
        chromsize_dict = dict(zip(chromsizes[CHROM_COL], chromsizes[END_COL], strict=True))

    gr = ensure_pyranges(chromsizes).sort_ranges().tile_ranges(tile_size).reset_index(drop=True)

    # truncate the last tile when requested
    if not full_last_tile:
        # map each row to its chromosome length
        chrom_ends = gr[CHROM_COL].map(chromsize_dict)

        # clip only those End values that run past the chromosome length
        gr.loc[:, END_COL] = np.minimum(gr[END_COL].to_numpy(), chrom_ends.to_numpy())

    return ensure_pyranges(gr)



import pyranges as pr


def test_mcc() -> None:
    gr1, gr2 = pr.example_data.chipseq(), pr.example_data.chipseq_background()
    g = pr.example_data.chromsizes()
    pyranges.ext.statistics.stats.mcc([gr1, gr2], genome=g, labels=["chip", "bg"], strand=True)



import typing
from dataclasses import dataclass

import numpy as np
import pandas as pd

if typing.TYPE_CHECKING:
    from pyranges import PyRanges
from pyranges.core.names import CHROM_COL, END_COL, START_COL, STRAND_COL, VALID_GENOMIC_STRAND_INFO
from pyranges.core.pyranges_helpers import ensure_pyranges

# Three types of accessors:
# 1. Columns or columns, rows
# 2. Genomic location


LociKeyType = str | int | slice | tuple[str | int, str] | tuple[str | int, slice] | tuple[str | int, str, slice]


@dataclass
class LociGetter:
    pr: "PyRanges"

    def __getitem__(
        self,
        key: LociKeyType,
    ) -> "PyRanges":
        return ensure_pyranges(self.pr.loc[self._matching_rows(key)])

    def _matching_rows(
        self,
        key: LociKeyType,
    ) -> "pd.Series[bool]":
        if isinstance(key, tuple):
            if is_chrom_and_strand(key):
                rows = chrom_and_strand(self.pr, key)
            elif is_chrom_or_strand_with_slice(key):
                rows = chrom_or_strand_with_slice(self.pr, key)
            elif is_3_tuple(key):
                rows = get_chrom_strand_and_range(self.pr, key)
            else:
                msg = f"Indexing tuple for loci must be of length 2 or 3, but was {len(key)}."
                raise ValueError(msg)
        elif isinstance(key, slice):
            rows = _rows_matching_range(self.pr, key)
        elif not isinstance(key, list):
            rows = get_chrom_and_strand(self.pr, key)
        else:
            msg = (
                "The loci accessor does not accept a list. If you meant to retrieve columns, use "
                "get_with_loc_columns instead."
            )
            raise TypeError(msg)
        return rows

    def __setitem__(self, key: typing.Any, value: typing.Any) -> None:
        rows = self._matching_rows(key)
        self.pr.loc[rows] = value


def _rows_matching_chrom(gr: "PyRanges", chrom: str | int | None) -> "pd.Series[bool]":
    return (gr[CHROM_COL].astype(type(chrom)) == chrom) if chrom is not None else pd.Series(data=True, index=gr.index)


def _rows_matching_strand(gr: "PyRanges", strand: str | None) -> "pd.Series[bool]":
    return (gr[STRAND_COL] == strand) if strand is not None else pd.Series(data=True, index=gr.index)


def _rows_matching_range(gr: "PyRanges", _range: slice | None) -> "pd.Series[bool]":
    if _range is None:
        return pd.Series(data=True, index=gr.index)
    start_in_range = gr[START_COL] < (_range.stop if _range.stop is not None else np.inf)
    end_in_range = gr[END_COL] > (_range.start if _range.start is not None else -1)
    return start_in_range & end_in_range


def _rows_matching_chrom_and_strand(gr: "PyRanges", chrom: str | int | None, strand: str | None) -> "pd.Series[bool]":
    return _rows_matching_chrom(gr, chrom) & _rows_matching_strand(gr, strand)


def _rows_matching_chrom_and_strand_and_range(
    gr: "PyRanges",
    chrom: str | None | int,
    strand: str | None,
    _range: slice | None,
) -> pd.Series:
    return _rows_matching_chrom(gr, chrom) & _rows_matching_strand(gr, strand) & _rows_matching_range(gr, _range)


def is_3_tuple(key: tuple) -> bool:
    """Check if key is a 3-tuple."""
    return len(key) == 3  # noqa: PLR2004


def is_2_tuple(key: tuple) -> bool:
    """Check if key is a 2-tuple."""
    return len(key) == 2  # noqa: PLR2004


def is_chrom_or_strand_with_slice(key: tuple) -> bool:
    """Check if key is a chromosome or strand and slice."""
    return is_2_tuple(key) and isinstance(key[1], slice)


def is_chrom_and_strand(key: tuple) -> bool:
    """Check if key is a chromosome and strand."""
    return is_2_tuple(key) and isinstance(key[1], str)


def chrom_and_strand(pr: "PyRanges", key: tuple) -> "pd.Series[bool]":
    """Get rows matching chromosome and strand."""
    chrom, strand = key
    return _rows_matching_chrom_and_strand(pr, chrom, strand)


def chrom_or_strand_with_slice(pr: "PyRanges", key: tuple) -> "pd.Series[bool]":
    """Get rows matching chromosome or strand and slice."""
    chrom_or_strand, loc = key

    # We can get a view with right chromosome, then match range only there, and reindex to original.
    # Or we can interrogate the whole PyRanges for chromosome and range separately, then combine the boolean Series
    # I timeit tested random pyranges up to 10^8 rows.
    # The second was consistently faster, so here it is:

    # in ambyguous cases, we test is strand is + or -, otherwise we assume it's a chromosome
    if chrom_or_strand in VALID_GENOMIC_STRAND_INFO and pr.has_strand:
        rows = _rows_matching_strand(pr, chrom_or_strand) & (_rows_matching_range(pr, loc))
    else:
        rows = _rows_matching_chrom(pr, chrom_or_strand) & (_rows_matching_range(pr, loc))
    return rows


def get_chrom_strand_and_range(pr: "PyRanges", key: tuple) -> "pd.Series":
    """Get rows matching chromosome, strand and range."""
    chrom, strand, _range = key
    return _rows_matching_chrom_and_strand_and_range(pr, chrom, strand, _range)


def get_chrom_and_strand(pr: "PyRanges", key: str | int) -> "pd.Series[bool]":
    """Get rows matching chromosome or strand.

    We do not know whether the key is a chromosome or a strand.
    We determine it's strand only if + or -, chromosome otherwise
    """
    if key in VALID_GENOMIC_STRAND_INFO and pr.has_strand:
        rows = _rows_matching_strand(pr, key)
    else:
        rows = _rows_matching_chrom(pr, key)
    return rows



from itertools import product

import numpy as np
import pytest
from hypothesis import HealthCheck, given, settings

from tests.property_based.hypothesis_helper import deadline, dfs_no_min, max_examples

strandedness = [False, "same", "opposite"]

binary_methods = [
    "set_union",
    "set_intersect",
    "overlap",
    "nearest",
    "intersect",
    "subtract",
    "join",
]

unary_methods = [
    "merge",
    "sort",
    "cluster",
    "pc",
    "mpc",
    "spc",
    "drop_duplicate_positions",
    "drop",
]

method_chain = product(binary_methods, binary_methods)
# cannot start with an operation that makes pyrange unstranded and then try a stranded op
strandedness_chain = list(product(["same", "opposite"], strandedness)) + list(product(strandedness, [None]))


@pytest.mark.bedtools
@pytest.mark.parametrize(("strandedness_chain", "method_chain"), product(strandedness_chain, method_chain))
@settings(
    max_examples=max_examples,
    deadline=deadline,
    print_blob=True,
    suppress_health_check=HealthCheck.all(),
)
@given(gr=dfs_no_min(), gr2=dfs_no_min(), gr3=dfs_no_min())  # pylint: disable=no-value-for-parameter
# @reproduce_failure('5.5.4', b'AXicY2RAA4xIJCoLygcAALIABg==') # test_three_in_a_row[strandedness_chain122-method_chain122]
# @reproduce_failure('5.5.4', b'AXicY2QAAUYGKGBkxM9nAAABEAAJ') # test_three_in_a_row[strandedness_chain45-method_chain45]
# @reproduce_failure('5.5.4', b'AXicY2RAA4xIJDY+AAC2AAY=') # test_three_in_a_row[strandedness_chain24-method_chain24]
def test_three_in_a_row(gr, gr2, gr3, strandedness_chain, method_chain) -> None:
    s1, s2 = strandedness_chain
    f1, f2 = method_chain

    suffix_methods = ["nearest", "join"]

    if f1 in suffix_methods and f2 in suffix_methods:
        m1 = getattr(gr, f1)
        gr2 = m1(gr2, strandedness=s1)
        if len(gr2) > 0:
            assert gr2.Start.dtype == np.int64
            assert (gr2.Start >= 0).all()
            assert (gr2.End >= 0).all()
        m2 = getattr(gr2, f2)
        gr3 = m2(gr3, strandedness=s2, suffix="_c")
        if len(gr3) > 0:
            assert gr3.Start.dtype == np.int64
            assert (gr3.Start >= 0).all()
            assert (gr3.End >= 0).all()

    else:
        m1 = getattr(gr, f1)
        gr2 = m1(gr2, strandedness=s1)
        if len(gr2) > 0:
            assert gr2.Start.dtype == np.int64
            assert (gr2.Start >= 0).all()
            assert (gr2.End >= 0).all()
        m2 = getattr(gr2, f2)
        gr3 = m2(gr3, strandedness=s2)
        if len(gr3) > 0:
            assert gr3.Start.dtype == np.int64
            assert (gr3.Start >= 0).all()
            assert (gr3.End >= 0).all()


# @pytest.mark.bedtools
# @pytest.mark.parametrize("strandedness_chain,method_chain",
#                          product(strandedness_chain, method_chain))
# @settings(
#     max_examples=max_examples,
#     deadline=deadline,
#     suppress_health_check=HealthCheck.all())
# @given(gr=dfs_no_min(), gr2=dfs_no_min(), gr3=dfs_no_min())  # pylint: disable=no-value-for-parameter
# def test_three_in_a_row(gr, gr2, gr3, strandedness_chain, method_chain):

#     s1, s2 = strandedness_chain
#     f1, f2 = method_chain

#     # print(s1, s2)
#     # print(f1, f2)

#     m1 = getattr(gr, f1)
#     gr2 = m1(gr2, strandedness=s1)
#     m2 = getattr(gr2, f2)
#     gr3 = m2(gr3, strandedness=s2)



import pandas as pd


def assert_df_equal(df1, df2) -> None:
    # df1.loc[:, "Start"] = df1.Start.astype(np.int64)
    # df2.loc[:, "Start"] = df1.Start.astype(np.int64)
    # df1.loc[:, "End"] = df1.End.astype(np.int64)
    # df2.loc[:, "End"] = df1.End.astype(np.int64)

    pd.options.mode.chained_assignment = None
    if "Strand" in df1 and "Strand" in df2:
        sort_on = ["Chromosome", "Start", "End", "Strand"]
        df1.Strand = df1.Strand.astype("object")
        df2.Strand = df2.Strand.astype("object")
    else:
        sort_on = ["Chromosome", "Start", "End"]

    if "Strand_b" in df1:
        sort_on += ["Start_b", "End_b", "Strand_b"]
        df1.Strand_b = df1.Strand_b.astype("object")
        df2.Strand_b = df2.Strand_b.astype("object")
    elif "Start_b" in df2:
        sort_on += ["Start_b", "End_b"]

    df1 = df1.sort_values(sort_on)
    df2 = df2.sort_values(sort_on)

    df1 = df1.reset_index(drop=True)
    df2 = df2.reset_index(drop=True)

    df1.Chromosome = df1.Chromosome.astype("object")
    df2.Chromosome = df2.Chromosome.astype("object")

    # print("dtypes Strand\n", "1",  df1.Strand.dtype, "2", df2.Strand.dtype)
    # print("dtypes Strand\n", df1.Strand.dtype == df2.Strand.dtype)
    # print("dtypes equal\n", df1.dtypes == df2.dtypes)

    pd.testing.assert_frame_equal(df1, df2)

    pd.options.mode.chained_assignment = "warn"



from typing import TYPE_CHECKING, Any

import numpy as np
import pandas as pd

from pyranges.core.names import CHROM_COL, END_COL, START_COL
from pyranges.core.pyranges_helpers import ensure_rangeframe

if TYPE_CHECKING:
    from pyranges.range_frame.range_frame import RangeFrame


def _complement(
    df: "RangeFrame",
    *,
    by: list[str],
    slack: int = 0,
    chromsizes_col: str | None = None,
    chromsizes: "dict[str | int, int] | None" = None,
    include_first_interval: bool = False,
) -> "RangeFrame":
    import ruranges

    from pyranges.range_frame.range_frame import RangeFrame

    if df.empty:
        return df

    col_order = [col for col in df if col in [*by, START_COL, END_COL]]

    factorized = pd.Series(np.zeros(len(df), dtype=np.uint32)) if not by else df.groupby(by).ngroup().astype(np.uint32)
    # The Rust kernel must see the *same* dtype that we pass for starts/ends
    pos_dtype: np.dtype[Any] = df[START_COL].to_numpy(copy=False).dtype
    grp_dtype: np.dtype[Any] = factorized.to_numpy(copy=False).dtype

    if chromsizes and chromsizes_col:
        the_chromsizes_col = df[chromsizes_col]
        if isinstance(the_chromsizes_col.dtype, pd.CategoricalDtype):
            # drop the categorical metadata; cheap view, no copy
            the_chromsizes_col = the_chromsizes_col.astype(object)

        #  vectorised lookup with no FutureWarning
        lengths = (
            the_chromsizes_col.map(chromsizes).astype(  # faster/cleaner than replace
                pos_dtype, copy=False
            )  # ensure same dtype as starts/ends
        )

        group_to_len = pd.DataFrame(
            {
                "group_id": factorized,
                "length": lengths,
            }
        ).drop_duplicates()
        chrom_len_ids = group_to_len["group_id"].to_numpy(grp_dtype)
        chrom_lens = group_to_len["length"].to_numpy(pos_dtype)

    else:
        chrom_len_ids = np.array([], dtype=grp_dtype)
        chrom_lens = np.array([], dtype=pos_dtype)

    chrs, start, end, idxs = ruranges.complement(
        groups=factorized.to_numpy(),
        starts=df.Start.to_numpy(),
        ends=df.End.to_numpy(),
        slack=slack,
        chrom_len_ids=chrom_len_ids,  # type: ignore[arg-type]
        chrom_lens=chrom_lens,
        include_first_interval=include_first_interval,
    )

    ids = df.take(idxs)  # type: ignore[arg-type]

    result = RangeFrame({CHROM_COL: chrs, START_COL: start, END_COL: end} | {_by: ids[_by] for _by in by})[col_order]

    return ensure_rangeframe(result.reset_index(drop=True))
